import numpy as np

from gpaw.mpi import have_mpi
from gpaw.utilities import compiled_with_libvdwxc
from gpaw.utilities.grid_redistribute import Domains, general_redistribute
from gpaw.utilities.timing import nulltimer
from gpaw.xc.functional import XCFunctional
from gpaw.xc.gga import GGA, gga_vars, add_gradient_correction
from gpaw.xc.libxc import LibXC
from gpaw.xc.mgga import MGGA
import gpaw
import _gpaw


def libvdwxc_has_mpi():
    return have_mpi and _gpaw.libvdwxc_has('mpi')


def libvdwxc_has_pfft():
    return have_mpi and _gpaw.libvdwxc_has('pfft')


def check_grid_descriptor(gd):
    assert gd.parsize_c[1] == 1 and gd.parsize_c[2] == 1
    nxpts_p = gd.n_cp[0][1:] - gd.n_cp[0][:-1]
    nxpts0 = nxpts_p[0]
    for nxpts in nxpts_p[1:-1]:
        assert nxpts == nxpts0
    assert nxpts_p[-1] <= nxpts0


def get_domains(N_c, parsize_c):
    # We want a distribution like this:
    #   [B, B, ..., B, remainder, 0, 0, ..., 0].
    # with blocksize B chosen as large as possible for better load balance.
    # This function returns the blocksize and the cumulative sum of indices
    # starting with 0.
    blocksize_c = -(-N_c // parsize_c)
    return (np.arange(1 + parsize_c) * blocksize_c).clip(0, N_c)


def get_auto_pfft_grid(size):
    nproc1 = size
    nproc2 = 1
    while nproc1 > nproc2 and nproc1 % 2 == 0:
        nproc1 //= 2
        nproc2 *= 2
    return nproc1, nproc2


spinwarning = """\
GPAW uses the total density to evaluate the van der Waals functional
for a spin-polarized system.  This is not entirely rigorous, so the
calculation cannot be considered a true vdW-DF-family calculation."""
_VDW_NUMERICAL_CODES = {'vdW-DF': 1,
                        'vdW-DF2': 2,
                        'vdW-DF-cx': 3}


class LibVDWXC(object):
    """Minimum-tomfoolery object-oriented interface to libvdwxc."""
    def __init__(self, funcname, N_c, cell_cv, comm, mode='auto',
                 pfft_grid=None, nspins=1):
        self.initialized = False
        if not compiled_with_libvdwxc():
            raise ImportError('libvdwxc not compiled into GPAW')

        self.vdw_functional_name = funcname
        code = _VDW_NUMERICAL_CODES[funcname]
        self.shape = tuple(N_c)
        ptr = np.empty(1, np.intp)
        _gpaw.libvdwxc_create(ptr, code, nspins, self.shape,
                              tuple(np.ravel(cell_cv)))
        # assign ptr only now that it is initialized (so __del__ always works)
        self._ptr = ptr

        # Choose mode automatically if necessary:
        if mode == 'auto':
            if pfft_grid is not None:
                mode = 'pfft'
            elif comm.size > 1:
                mode = 'mpi'
            else:
                mode = 'serial'
        assert mode in ['serial', 'mpi', 'pfft']

        if mode != 'serial' and not have_mpi:
            raise ImportError('MPI not available for libvdwxc-%s '
                              'because GPAW is serial' % mode)

        if mode == 'pfft':
            if pfft_grid is None:
                pfft_grid = get_auto_pfft_grid(comm.size)
            nx, ny = pfft_grid
            assert nx * ny == comm.size
            # User might have passed a list, but we make sure to store a tuple:
            self.pfft_grid = (nx, ny)
        elif pfft_grid is not None:
            raise ValueError('pfft_grid specified with mode %s' % mode)

        self.mode = mode
        self.comm = comm

    def initialize_backend(self):
        assert not self.initialized

        mode = self.mode
        comm = self.comm
        if mode == 'serial':
            assert comm.size == 1, ('You cannot run in serial with %d cores'
                                    % comm.size)
            _gpaw.libvdwxc_init_serial(self._ptr)
        elif gpaw.dry_run:
            pass
            # XXXXXX liable to cause really nasty errors.
        elif mode == 'mpi':
            if not libvdwxc_has_mpi():
                raise ImportError('libvdwxc not compiled with MPI')
            _gpaw.libvdwxc_init_mpi(self._ptr, comm.get_c_object())
        elif mode == 'pfft':
            nx, ny = self.pfft_grid
            _gpaw.libvdwxc_init_pfft(self._ptr, comm.get_c_object(), nx, ny)

        self.initialized = True

    def calculate(self, n_sg, sigma_xg, dedn_sg, dedsigma_xg):
        """Calculate energy and add partial derivatives to arrays."""
        if not self.initialized:
            self.initialize_backend()

        for arr in [n_sg, sigma_xg, dedn_sg, dedsigma_xg]:
            assert arr.flags.contiguous
            assert arr.dtype == float
            # XXX We cannot actually ask libvdwxc about its expected shape
            # assert arr.shape == self.shape, [arr.shape, self.shape]
        energy = _gpaw.libvdwxc_calculate(self._ptr, n_sg, sigma_xg,
                                          dedn_sg, dedsigma_xg)
        return energy

    def get_description(self):
        if self.mode == 'serial':
            pardesc = 'fftw3 serial'
        elif self.mode == 'mpi':
            size = self.comm.size
            cores = '%d cores' % size if size != 1 else 'one core'
            pardesc = 'fftw3-mpi with %s' % cores
        else:
            assert self.mode == 'pfft'
            pardesc = 'pfft with %d x %d CPU grid' % self.pfft_grid
        return '%s [libvdwxc/%s]' % (self.vdw_functional_name, pardesc)

    def tostring(self):
        return _gpaw.libvdwxc_tostring(self._ptr)

    def __del__(self):
        if hasattr(self, '_ptr'):
            _gpaw.libvdwxc_free(self._ptr)


class RedistWrapper:
    """Call libvdwxc redistributing automatically from and to GPAW grid."""
    def __init__(self, libvdwxc, distribution, timer=nulltimer,
                 vdwcoef=1.0):
        # It is hacky for the RedistWrapper to apply the vdwcoef, but this
        # is the only accessible place where we take copies of the arrays,
        # and therefore the only 'good' place to apply a factor without
        # applying it to the existing contents of the array.
        self.libvdwxc = libvdwxc
        self.distribution = distribution
        self.timer = timer
        self.vdwcoef = vdwcoef

    def calculate(self, n_sg, sigma_xg, v_sg, dedsigma_xg):
        zeros = self.distribution.block_zeros
        sshape = (len(n_sg),)
        xshape = (len(sigma_xg),)
        nblock_sg = zeros(sshape)
        sigmablock_xg = zeros(xshape)
        vblock_sg = zeros(sshape)
        dedsigmablock_xg = zeros(xshape)

        self.timer.start('redistribute')
        self.distribution.gd2block(n_sg, nblock_sg)
        self.distribution.gd2block(sigma_xg, sigmablock_xg)
        self.timer.stop('redistribute')

        self.timer.start('libvdwxc nonlocal')
        energy = self.libvdwxc.calculate(nblock_sg, sigmablock_xg,
                                         vblock_sg, dedsigmablock_xg)
        self.timer.stop('libvdwxc nonlocal')
        energy *= self.vdwcoef
        for arr in vblock_sg, dedsigmablock_xg:
            arr *= self.vdwcoef

        self.timer.start('redistribute')
        self.distribution.block2gd_add(vblock_sg, v_sg)
        self.distribution.block2gd_add(dedsigmablock_xg, dedsigma_xg)
        self.timer.stop('redistribute')
        return energy


class FFTDistribution:
    def __init__(self, gd, parsize_c):
        self.input_gd = gd
        assert np.prod(parsize_c) == gd.comm.size
        self.local_input_size_c = gd.n_c
        self.domains_in = Domains(gd.n_cp)
        N_c = gd.get_size_of_global_array(pad=True)
        self.domains_out = Domains([get_domains(N_c[i], parsize_c[i])
                                    for i in range(3)])

        if parsize_c[1] == 1 and parsize_c[2] == 1:
            def aux_rank_to_parpos(rank=gd.comm.rank):
                return np.array([rank, 0, 0], int)
        else:
            # For 2D distributions we use a grid descriptor.  We could
            # actually use a grid descriptor always, but that causes trouble
            # when there are more cores than grid points, which could well
            # be the case when the distribution is only 1D.
            #
            # The auxiliary gd actually is used *only* for the rank/parpos
            # correspondence.  The actual domains it defines are unused!!
            aux_gd = gd.new_descriptor(comm=gd.comm, parsize_c=parsize_c)
            aux_rank_to_parpos = aux_gd.get_processor_position_from_rank

        parpos_c = aux_rank_to_parpos()
        self.aux_rank_to_parpos = aux_rank_to_parpos
        self.local_output_size_c = tuple(self.domains_out.get_box(parpos_c)[1])

    def block_zeros(self, shape=()):
        return np.zeros(shape + self.local_output_size_c)

    def gd2block(self, a_xg, b_xg):
        general_redistribute(self.input_gd.comm,
                             self.domains_in, self.domains_out,
                             self.input_gd.get_processor_position_from_rank,
                             self.aux_rank_to_parpos,
                             a_xg, b_xg, behavior='overwrite')

    def block2gd_add(self, a_xg, b_xg):
        general_redistribute(self.input_gd.comm,
                             self.domains_out, self.domains_in,
                             self.aux_rank_to_parpos,
                             self.input_gd.get_processor_position_from_rank,
                             a_xg, b_xg, behavior='add')


class VDWXC(XCFunctional):
    def __init__(self, semilocal_xc, name, mode='auto',
                 pfft_grid=None, libvdwxc_name=None,
                 setup_name='revPBE', vdwcoef=1.0,
                 accept_partial_decomposition=False):
        """Initialize VDWXC object (further initialization required).

        mode can be 'auto', 'serial', 'mpi', or 'pfft'.

         * 'serial' uses FFTW and only works with serial decompositions.

         * 'mpi' uses FFTW-MPI with communicator of the grid
           descriptor, parallelizing along the x axis.

         * 'pfft' uses PFFT and works with any decomposition,
           parallelizing along two directions for best scalability.

         * 'auto' uses PFFT if pfft_grid is given, else FFTW-MPI if the
           calculation uses more than one core, else serial FFTW.

         pfft_grid is the 2D CPU grid used by PFFT and can be a tuple
         (nproc1, nproc2) that multiplies to total communicator size,
         or None.  It is an error to specify pfft_grid unless using
         PFFT.  If left unspecified, a hopefully reasonable automatic
         choice will be made.
         """

        # XXX We should probably not initialize with the same data as the
        # semilocal XC kernel.
        XCFunctional.__init__(self, semilocal_xc.kernel.name,
                              semilocal_xc.kernel.type)
        # Really, 'type' should be something along the lines of vdw-df.
        self.semilocal_xc = semilocal_xc

        # We set these in the initialize later (ugly).
        self.libvdwxc = None
        self.distribution = None
        self.redist_wrapper = None
        self.timer = nulltimer
        self.setup_name = setup_name

        # These parameters are simply stored for later forwarding
        self.name = name
        if libvdwxc_name is None:
            libvdwxc_name = name
        self._libvdwxc_name = libvdwxc_name
        self._mode = mode
        self._pfft_grid = pfft_grid
        self._vdwcoef = vdwcoef
        self.accept_partial_decomposition = accept_partial_decomposition
        self._nspins = 1

        self.last_nonlocal_energy = None
        self.last_semilocal_energy = None

        # XXXXXXXXXXXXXXXXX
        self.calculate_paw_correction = semilocal_xc.calculate_paw_correction
        self.calculate_spherical = semilocal_xc.calculate_spherical
        self.apply_orbital_dependent_hamiltonian = \
            semilocal_xc.apply_orbital_dependent_hamiltonian
        self.add_forces = semilocal_xc.add_forces
        self.get_kinetic_energy_correction = \
            semilocal_xc.get_kinetic_energy_correction
        self.rotate = semilocal_xc.rotate

    def __str__(self):
        tokens = [self._mode]
        if self._libvdwxc_name != self.name:
            tokens.append('nonlocal-name={0}'.format(self._libvdwxc_name))
            tokens.append('gga-kernel={0}'
                          .format(self.semilocal_xc.kernel.name))
        if self._pfft_grid is not None:
            tokens.append('pfft={0}'.format(self._pfft_grid))
        if self._vdwcoef != 1.0:
            tokens.append('vdwcoef={0}'.format(self._vdwcoef))

        qualifier = ', '.join(tokens)
        return '{0} [libvdwxc/{1}]'.format(self.name, qualifier)

    def todict(self):
        dct = dict(backend='libvdwxc',
                   semilocal_xc=self.semilocal_xc.name,
                   name=self.name,
                   # mode=self._mode,
                   # pfft_grid=self._pfft_grid,
                   libvdwxc_name=self._libvdwxc_name,
                   setup_name=self.setup_name,
                   vdwcoef=self._vdwcoef)
        return dct

    def set_grid_descriptor(self, gd):
        XCFunctional.set_grid_descriptor(self, gd)
        self.semilocal_xc.set_grid_descriptor(gd)

    def get_description(self):
        lines = []
        app = lines.append
        app('{} with libvdwxc'.format(self.name))
        mode = self.libvdwxc.mode
        ncores = self.libvdwxc.comm.size
        cores = 'core' if ncores == 1 else 'cores'
        if mode == 'mpi':
            mode = 'mpi with {} {}'.format(ncores, cores)
        elif mode == 'pfft':
            nx, ny = self.libvdwxc.pfft_grid
            mode = 'pfft with {} x {} {}'.format(nx, ny, cores)
        app('Mode: {}'.format(mode))
        app('Semilocal: {}'.format(self.semilocal_xc.get_description()))
        if self.libvdwxc.vdw_functional_name != self.name:
            app('Corresponding non-local functional: {}'
                .format(self.libvdwxc.vdw_functional_name))
        app('Local blocksize: {} x {} x {}'
            .format(*self.distribution.local_output_size_c))
        app('PAW datasets: {}'.format(self.get_setup_name()))
        return '\n'.join(lines)

    def summary(self, log):
        from ase.units import Hartree
        enl = self.libvdwxc.comm.sum(self.last_nonlocal_energy)
        esl = self.gd.comm.sum(self.last_semilocal_energy)
        # In the current implementation these communicators have the same
        # processes always:
        assert self.libvdwxc.comm.size == self.gd.comm.size
        log('Non-local %s correlation energy: %.6f' % (self.name,
                                                       enl * Hartree))
        log('Semilocal %s energy: %.6f' % (self.semilocal_xc.kernel.name,
                                           esl * Hartree))
        log('(Not including atomic contributions)')

    def get_setup_name(self):
        return self.setup_name

    def initialize_backend(self, gd):
        N_c = gd.get_size_of_global_array(pad=True)
        # garbage-collect before allocating next one, if an object was
        # already allocated:
        self.libvdwxc = None
        self.libvdwxc = LibVDWXC(self._libvdwxc_name, N_c, gd.cell_cv,
                                 gd.comm, mode=self._mode,
                                 pfft_grid=self._pfft_grid,
                                 nspins=self._nspins)
        cpugrid = [1, 1, 1]
        if self.libvdwxc.mode == 'mpi':
            cpugrid[0] = gd.comm.size
        elif self.libvdwxc.mode == 'pfft':
            cpugrid[0], cpugrid[1] = self.libvdwxc.pfft_grid
        self.distribution = FFTDistribution(gd, cpugrid)
        self.redist_wrapper = RedistWrapper(self.libvdwxc,
                                            self.distribution,
                                            self.timer,
                                            self._vdwcoef)

    def set_positions(self, spos_ac):
        self.semilocal_xc.set_positions(spos_ac)

    def initialize(self, density, hamiltonian, wfs):
        self.timer = hamiltonian.timer  # fragile object robbery
        self.semilocal_xc.initialize(density, hamiltonian, wfs)
        gd = self.gd
        self._nspins = density.nspins

        self.initialize_backend(self.gd)

        if (wfs.world.size > gd.comm.size and np.prod(gd.N_c) > 64**3
            and not self.accept_partial_decomposition):
            # We could issue a warning if an excuse turns out to exist some day
            raise ValueError('You are using libvdwxc with only '
                             '%d out of %d available cores in a non-small '
                             'calculation (%s points).  This is not '
                             'a crime but is likely silly and therefore '
                             'triggers an error.  Please use '
                             'parallel={\'augment_grids\': True} '
                             'or complain to the developers.  '
                             'You can also disable this error '
                             'by passing accept_partial_decomposition=True '
                             'to the libvdwxc functional object, '
                             'but be careful to use good domain '
                             'decomposition.' %
                             (gd.comm.size, wfs.world.size,
                              ' x '.join(str(N) for N in gd.N_c)))
        # TODO Here we could decide FFT padding.

    def calculate_impl(self, gd, n_sg, v_sg, e_g):
        """Calculate energy and potential.

        gd may be non-periodic.  To be distinguished from self.gd
        which is always periodic due to priminess of FFT dimensions.
        (To do: proper padded FFTs.)"""
        assert gd == self.distribution.input_gd
        assert self.libvdwxc is not None
        semiloc = self.semilocal_xc

        self.timer.start('van der Waals')

        self.timer.start('semilocal')
        # XXXXXXX taken from GGA
        grad_v = semiloc.grad_v
        sigma_xg, dedsigma_xg, gradn_svg = gga_vars(gd, grad_v, n_sg)
        n_sg[:] = np.abs(n_sg)  # XXXX What to do about this?
        sigma_xg[:] = np.abs(sigma_xg)

        # Grrr, interface still sucks
        if hasattr(semiloc, 'process_mgga'):
            semiloc.process_mgga(e_g, n_sg, v_sg, sigma_xg, dedsigma_xg)
        else:
            semiloc.kernel.calculate(e_g, n_sg, v_sg, sigma_xg, dedsigma_xg)
        self.last_semilocal_energy = e_g.sum() * self.gd.dv
        self.timer.stop('semilocal')

        energy_nonlocal = self.redist_wrapper.calculate(n_sg, sigma_xg,
                                                        v_sg, dedsigma_xg)
        # Note: Redistwrapper handles vdwcoef.  For now

        add_gradient_correction(grad_v, gradn_svg, sigma_xg, dedsigma_xg, v_sg)

        self.last_nonlocal_energy = energy_nonlocal
        e_g[0, 0, 0] += energy_nonlocal / self.gd.dv
        self.timer.stop('van der Waals')


def vdw_df(**kwargs):
    kwargs1 = dict(name='vdW-DF', setup_name='revPBE',
                   semilocal_xc=GGA(LibXC('GGA_X_PBE_R+LDA_C_PW'),
                                    stencil=kwargs.pop('stencil', 2)))
    kwargs1.update(kwargs)
    return VDWXC(**kwargs1)


def vdw_df2(**kwargs):
    kwargs1 = dict(name='vdW-DF2', setup_name='PBE',
                   semilocal_xc=GGA(LibXC('GGA_X_RPW86+LDA_C_PW'),
                                    stencil=kwargs.pop('stencil', 2)))
    kwargs1.update(kwargs)
    return VDWXC(**kwargs1)


def vdw_df_cx(**kwargs):
    # cx semilocal exchange is in libxc 2.2.2 or newer (or maybe from older)
    kernel = kwargs.get('kernel')
    if kernel is None:
        kernel = LibXC('GGA_X_LV_RPW86+LDA_C_PW')

    kwargs1 = dict(name='vdW-DF-cx', setup_name='PBE',
                   # PBEsol is most correct but not distributed by default.
                   semilocal_xc=GGA(kernel, stencil=kwargs.pop('stencil', 2)))
    kwargs1.update(kwargs)
    return VDWXC(**kwargs1)


def vdw_optPBE(**kwargs):
    kwargs1 = dict(name='vdW-optPBE', libvdwxc_name='vdW-DF', setup_name='PBE',
                   semilocal_xc=GGA(LibXC('GGA_X_OPTPBE_VDW+LDA_C_PW'),
                                    stencil=kwargs.pop('stencil', 2)))
    kwargs1.update(kwargs)
    return VDWXC(**kwargs1)


def vdw_optB88(**kwargs):
    kwargs1 = dict(name='optB88', libvdwxc_name='vdW-DF', setup_name='PBE',
                   semilocal_xc=GGA(LibXC('GGA_X_OPTB88_VDW+LDA_C_PW'),
                                    stencil=kwargs.pop('stencil', 2)))
    kwargs1.update(kwargs)
    return VDWXC(**kwargs1)


def vdw_C09(**kwargs):
    kwargs1 = dict(name='vdW-C09', libvdwxc_name='vdW-DF', setup_name='PBE',
                   semilocal_xc=GGA(LibXC('GGA_X_C09X+LDA_C_PW'),
                                    stencil=kwargs.pop('stencil', 2)))
    kwargs1.update(kwargs)
    return VDWXC(**kwargs1)


def vdw_beef(**kwargs):
    # Kernel parameters stolen from vdw.py
    from gpaw.xc.bee import BEEVDWKernel
    kernel = BEEVDWKernel('BEE2', None,
                          0.600166476948828631066,
                          0.399833523051171368934)
    kwargs1 = dict(name='vdW-BEEF', libvdwxc_name='vdW-DF2', setup_name='PBE',
                   semilocal_xc=GGA(kernel, stencil=kwargs.pop('stencil', 2)))
    kwargs1.update(kwargs)
    return VDWXC(**kwargs1)


def vdw_mbeef(**kwargs):
    # Note: Parameters taken from vdw.py
    from gpaw.xc.bee import BEEVDWKernel
    kernel = BEEVDWKernel('BEE3', None, 0.405258352, 0.356642240)
    kwargs1 = dict(name='vdW-mBEEF', setup_name='PBEsol',
                   libvdwxc_name='vdW-DF2', vdwcoef=0.886774972,
                   semilocal_xc=MGGA(kernel, stencil=kwargs.pop('stencil', 2)))
    kwargs1.update(kwargs)
    return VDWXC(**kwargs1)


# String to functional mapping
def get_libvdwxc_functional(name, **kwargs):
    if 'name' in kwargs:
        name2 = kwargs.pop('name')
        assert name == name2
    funcs = {'vdW-DF': vdw_df,
             'vdW-DF2': vdw_df2,
             'vdW-DF-cx': vdw_df_cx,
             'optPBE-vdW': vdw_optPBE,
             'optB88-vdW': vdw_optB88,
             'C09-vdW': vdw_C09,
             'BEEF-vdW': vdw_beef,
             'mBEEF-vdW': vdw_mbeef}

    semilocal_xc = kwargs.pop('semilocal_xc', None)
    if semilocal_xc is not None:
        from gpaw.xc import XC
        semilocal_xc = XC(semilocal_xc)

    func = funcs[name](**kwargs)
    if semilocal_xc is not None:
        assert semilocal_xc.name == func.semilocal_xc.name
    return func


class CXGGAKernel:
    def __init__(self, just_kidding=False):
        self.just_kidding = just_kidding
        self.type = 'GGA'
        self.lda_c = LibXC('LDA_C_PW')
        if self.just_kidding:
            self.name = 'purepython rPW86_with_%s' % self.lda_c.name
        else:
            self.name = 'purepython cx'

    def calculate(self, e_g, n_sg, v_sg, sigma_xg, dedsigma_xg):
        e_g[:] = 0.0
        dedsigma_xg[:] = 0.0

        self.lda_c.calculate(e_g, n_sg, v_sg, sigma_xg, dedsigma_xg)

        for arr in [n_sg, v_sg, sigma_xg, dedsigma_xg]:
            assert len(arr) == 1
        self._exchange(n_sg[0], sigma_xg[0], e_g, v_sg[0], dedsigma_xg[0])

    def _exchange(self, rho, grho, sx, v1x, v2x):
        """Calculate cx local exchange.

        Note that this *adds* to the energy density sx so that it can
        be called after LDA correlation part without ruining anything.
        Also it adds to v1x and v2x as is normal in GPAW."""
        tol = 1e-20
        rho[rho < tol] = tol
        grho[grho < tol] = tol
        alp = 0.021789
        beta = 1.15
        a = 1.851
        b = 17.33
        c = 0.163
        mu_LM = 0.09434
        s_prefactor = 6.18733545256027
        Ax = -0.738558766382022  # = -3./4. * (3./pi)**(1./3)
        four_thirds = 4. / 3.

        grad_rho = np.sqrt(grho)

        # eventually we need s to power 12.  Clip to avoid overflow
        # (We have individual tolerances on both rho and grho, but
        # they are not sufficient to guarantee this)
        s_1 = (grad_rho / (s_prefactor * rho**four_thirds)).clip(0.0, 1e20)
        s_2 = s_1 * s_1
        s_3 = s_2 * s_1
        s_4 = s_3 * s_1
        s_5 = s_4 * s_1
        s_6 = s_5 * s_1

        fs_rPW86 = (1.0 + a * s_2 + b * s_4 + c * s_6)**(1. / 15.)

        if self.just_kidding:
            fs = fs_rPW86
        else:
            fs = (1.0 + mu_LM * s_2) / (1.0 + alp * s_6) \
                + alp * s_6 / (beta + alp * s_6) * fs_rPW86

        # the energy density for the exchange.
        sx[:] += Ax * rho**four_thirds * fs

        df_rPW86_ds = (1. / (15. * fs_rPW86**14.0)) * \
            (2 * a * s_1 + 4 * b * s_3 + 6 * c * s_5)

        if self.just_kidding:
            df_ds = df_rPW86_ds  # XXXXXXXXXXXXXXXXXXXX
        else:
            df_ds = 1. / (1. + alp * s_6)**2 \
                * (2.0 * mu_LM * s_1 * (1. + alp * s_6) -
                   6.0 * alp * s_5 * (1. + mu_LM * s_2)) \
                + alp * s_6 / (beta + alp * s_6) * df_rPW86_ds \
                + 6.0 * alp * s_5 * fs_rPW86 / (beta + alp * s_6) \
                * (1. - alp * s_6 / (beta + alp * s_6))

        # de/dn.  This is the partial derivative of sx wrt. n, for s constant
        v1x[:] += Ax * four_thirds * (rho**(1. / 3.) * fs -
                                      grad_rho / (s_prefactor * rho) * df_ds)
        # de/d(nabla n).  The other partial derivative
        v2x[:] += 0.5 * Ax * df_ds / (s_prefactor * grad_rho)
        # (We may or may not understand what that grad_rho is doing here.)


def test_derivatives():
    gen = np.random.RandomState(1)
    shape = (1, 20, 20, 20)
    ngpts = np.prod(shape)
    n_sg = gen.rand(*shape)
    sigma_xg = np.zeros(shape)
    sigma_xg[:] = gen.rand(*shape)

    qe_kernel = CXGGAKernel(just_kidding=True)
    libxc_kernel = LibXC('GGA_X_RPW86+LDA_C_PW')

    cx_kernel = CXGGAKernel(just_kidding=False)

    def check(kernel, n_sg, sigma_xg):
        e_g = np.zeros(shape[1:])
        dedn_sg = np.zeros(shape)
        dedsigma_xg = np.zeros(shape)
        kernel.calculate(e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg)
        return e_g, dedn_sg, dedsigma_xg

    def check_and_write(kernel):
        n1_sg = n_sg.copy()
        e_g, dedn_sg, dedsigma_xg = check(kernel, n_sg, sigma_xg)
        dedn = dedn_sg[0, 0, 0, 0]
        dedsigma = dedsigma_xg[0, 0, 0, 0]

        dn = 1e-6
        n1_sg = n_sg.copy()
        n1_sg[0, 0, 0, 0] -= dn / 2.
        e1_g, _, _ = check(kernel, n1_sg, sigma_xg)

        n1_sg[0, 0, 0, 0] += dn
        e2_g, _, _ = check(kernel, n1_sg, sigma_xg)

        dedn_fd = (e2_g[0, 0, 0] - e1_g[0, 0, 0]) / dn
        dedn_err = abs(dedn - dedn_fd)

        print('e', e_g.sum() / ngpts)
        print('dedn', dedn, 'fd', dedn_fd, 'err %e' % dedn_err)

        sigma1_xg = sigma_xg.copy()
        sigma1_xg[0, 0, 0, 0] -= dn / 2.
        e1s_g, _, _ = check(kernel, n_sg, sigma1_xg)

        sigma1_xg[0, 0, 0, 0] += dn
        e2s_g, _, _ = check(kernel, n_sg, sigma1_xg)

        dedsigma_fd = (e2s_g[0, 0, 0] - e1s_g[0, 0, 0]) / dn
        dedsigma_err = dedsigma - dedsigma_fd

        print('dedsigma', dedsigma, 'fd', dedsigma_fd, 'err %e' % dedsigma_err)
        return e_g, dedn_sg, dedsigma_xg

    print('pw86r libxc')
    e_lxc_g, dedn_lxc_g, dedsigma_lxc_g = check_and_write(libxc_kernel)
    print()
    print('pw86r ours')
    e_qe_g, dedn_qe_g, dedsigma_qe_g = check_and_write(qe_kernel)
    print()
    print('cx')
    check_and_write(cx_kernel)

    print()
    print('lxc vs qe discrepancies')
    print('=======================')
    e_err = np.abs(e_lxc_g - e_qe_g).max()
    print('e', e_err)
    dedn_err = np.abs(dedn_qe_g - dedn_lxc_g).max()
    dedsigma_err = np.abs(dedsigma_lxc_g - dedsigma_qe_g).max()
    print('dedn', dedn_err)
    print('dedsigma', dedsigma_err)


def test_selfconsistent():
    from gpaw import GPAW
    from ase.build import molecule
    from gpaw.xc.gga import GGA

    system = molecule('H2O')
    system.center(vacuum=3.)

    def test(xc):
        calc = GPAW(mode='lcao',
                    xc=xc,
                    setups='sg15',
                    txt='gpaw.%s.txt' % str(xc)  # .kernel.name
                    )
        system.calc = calc
        return system.get_potential_energy()

    libxc_results = {}

    for name in ['GGA_X_PBE_R+LDA_C_PW', 'GGA_X_RPW86+LDA_C_PW']:
        xc = GGA(LibXC(name))
        e = test(xc)
        libxc_results[name] = e

    cx_gga_results = {}
    cx_gga_results['rpw86'] = test(GGA(CXGGAKernel(just_kidding=True)))
    cx_gga_results['lv_rpw86'] = test(GGA(CXGGAKernel(just_kidding=False)))

    vdw_results = {}
    vdw_coef0_results = {}

    for vdw in [vdw_df(), vdw_df2(), vdw_df_cx()]:
        vdw.vdwcoef = 0.0
        vdw_coef0_results[vdw.__class__.__name__] = test(vdw)
        vdw.vdwcoef = 1.0  # Leave nicest text file by running real calc last
        vdw_results[vdw.__class__.__name__] = test(vdw)

    from gpaw.mpi import world
    # These tests basically verify that the LDA/GGA parts of vdwdf
    # work correctly.
    if world.rank == 0:
        print('Now comparing...')
        err1 = cx_gga_results['rpw86'] - libxc_results['GGA_X_RPW86+LDA_C_PW']
        print('Our rpw86 must be identical to that of libxc. Err=%e' % err1)
        print('RPW86 interpolated with Langreth-Vosko stuff differs by %f'
              % (cx_gga_results['lv_rpw86'] - cx_gga_results['rpw86']))
        print('Each vdwdf with vdwcoef zero must yield same result as gga '
              'kernel')
        err_df1 = vdw_coef0_results['VDWDF'] - libxc_results['GGA_X_PBE_R+'
                                                             'LDA_C_PW']
        print('  df1 err=%e' % err_df1)
        err_df2 = vdw_coef0_results['VDWDF2'] - libxc_results['GGA_X_RPW86+'
                                                              'LDA_C_PW']
        print('  df2 err=%e' % err_df2)
        err_cx = vdw_coef0_results['VDWDFCX'] - cx_gga_results['lv_rpw86']
        print('   cx err=%e' % err_cx)


if __name__ == '__main__':
    test_derivatives()
    # test_selfconsistent()




from gpaw.utilities.tools import coordinates
from gpaw.utilities.gauss import Gaussian
from gpaw.grid_descriptor import GridDescriptor
from gpaw.helmholtz import HelmholtzSolver, ScreenedPoissonGaussian

# Initialize classes


def test_poisson_screened_poisson():
    a = 20  # Size of cell
    inv_width = 31  # inverse width of the gaussian
    N = 48  # Number of grid points
    coupling = -0.4  # dampening
    Nc = (N, N, N)                # Number of grid points along each axis
    gd = GridDescriptor(Nc, (a, a, a), 0)    # Grid-descriptor object
    solver = HelmholtzSolver(k2=coupling, nn=3)  # Numerical poisson solver
    # solver = PoissonSolver(nn=3)  # Numerical poisson solver
    # solver = HelmholtzSolver(0.16)  # Numerical poisson solver
    solver.set_grid_descriptor(gd)
    # Matrix with the square of the radial coordinate
    xyz, r2 = coordinates(gd)
    gauss = Gaussian(gd, a=inv_width)          # An instance of Gaussian
    test_screened_poisson = ScreenedPoissonGaussian(gd, a=inv_width)

    # Check if Gaussian potentials are made correctly

    # Array for storing the potential
    pot = gd.zeros(dtype=float, global_array=False)
    solver.load_gauss()
    vg = test_screened_poisson.get_phi(-coupling)  # esp. for dampening
    # Get analytic functions
    ng = gauss.get_gauss(0)
    #    vg = solver.phi_gauss
    # Solve potential numerically
    solver.solve(pot, ng, charge=None, zero_initial_phi=True)
    # Determine residual
    # residual = norm(pot - vg)
    residual = gd.integrate((pot - vg)**2)**0.5

    # print result
    print('residual %s' % (
        residual))
    assert residual < 0.003

    # mpirun -np 2 python gauss_func.py --gpaw-parallel --gpaw-debug



import atexit
import numpy as np
import _gpaw


def _elpaconstants():
    consts = _gpaw.pyelpa_constants()
    assert consts[0] == 0, 'ELPA_OK is {}, expected 0'.format(consts[0])
    return {'elpa_ok': consts[0],
            '1stage': consts[1],
            '2stage': consts[2]}


class LibElpa:
    @staticmethod
    def have_elpa():
        return hasattr(_gpaw, 'pyelpa_allocate')

    _elpa_initialized = False

    @staticmethod
    def api_version():
        return _gpaw.pyelpa_version()

    @classmethod
    def ensure_elpa_initialized(cls):
        if not cls._elpa_initialized:
            _gpaw.pyelpa_init()
            atexit.register(_gpaw.pyelpa_uninit)
            cls._elpa_initialized = True

    def __init__(self, desc, nev=None, solver='1stage'):
        if nev is None:
            nev = desc.gshape[0]

        ptr = np.zeros(1, np.intp)

        if not self.have_elpa():
            raise ImportError('GPAW is not running in parallel or otherwise '
                              'not compiled with Elpa support')

        if desc.nb != desc.mb:
            raise ValueError('Row and column block size must be '
                             'identical to support Elpa')

        self.ensure_elpa_initialized()
        _gpaw.pyelpa_allocate(ptr)
        self._ptr = ptr
        _gpaw.pyelpa_set_comm(ptr,
                              desc.blacsgrid.comm.get_c_object())
        self._parameters = {}

        self._consts = _elpaconstants()
        elpasolver = self._consts[solver]

        bg = desc.blacsgrid
        self.elpa_set(na=desc.gshape[0],
                      local_ncols=desc.shape[0],
                      local_nrows=desc.shape[1],
                      nblk=desc.mb,
                      process_col=bg.myrow,  # XXX interchanged
                      process_row=bg.mycol,
                      blacs_context=bg.context)
        self.elpa_set(nev=nev, solver=elpasolver)
        self.desc = desc

        _gpaw.pyelpa_setup(self._ptr)

    @property
    def description(self):
        solver = self._parameters['solver']
        if solver == self._consts['1stage']:
            pretty = 'Elpa one-stage solver'
        else:
            assert solver == self._consts['2stage']
            pretty = 'Elpa two-stage solver'
        return pretty

    @property
    def nev(self):
        return self._parameters['nev']

    def diagonalize(self, A, C, eps):
        assert self.nev == len(eps)
        self.desc.checkassert(A)
        self.desc.checkassert(C)
        _gpaw.pyelpa_diagonalize(self._ptr, A, C, eps)

    def general_diagonalize(self, A, S, C, eps, is_already_decomposed=0):
        assert self.nev == len(eps)
        self.desc.checkassert(A)
        self.desc.checkassert(S)
        self.desc.checkassert(C)
        _gpaw.pyelpa_general_diagonalize(self._ptr, A, S, C, eps,
                                         is_already_decomposed)

    def elpa_set(self, **kwargs):
        for key, value in kwargs.items():
            # print('pyelpa_set {}={}'.format(key, value))
            _gpaw.pyelpa_set(self._ptr, key, value)
            self._parameters[key] = value

    def __repr__(self):
        return 'LibElpa({})'.format(self._parameters)

    def __del__(self):
        if hasattr(self, '_ptr'):
            # elpa_deallocate has no error flag so we don't check it
            _gpaw.pyelpa_deallocate(self._ptr)
            self._ptr[0] = 0



import pytest

import numpy as np
from itertools import permutations

from gpaw.spherical_harmonics import Y
from gpaw.gaunt import gaunt, super_gaunt


def test_contraction_rule(lmax: int = 3):
    """Test that two spherical harmonics can be contracted to one."""
    G_LLL = gaunt(lmax)
    L1max, L2max, L3max = G_LLL.shape
    assert L2max == L3max

    x, y, z = unit_sphere_test_coordinates()
    for L1 in range(L1max):
        # In order to include all finite Gaunt coefficients in the expansion,
        # l1 + l2 has to be within the range of available L3 coefficients.
        l1 = int(np.sqrt(L1))
        l3max = int(np.sqrt(L3max)) - 1
        l2max = l3max - l1
        L2max = (l2max + 1)**2
        for L2 in range(L2max):
            product = Y(L1, x, y, z) * Y(L2, x, y, z)
            expansion = 0.
            for L3 in range(L3max):
                expansion += G_LLL[L1, L2, L3] * Y(L3, x, y, z)
            assert expansion == pytest.approx(product)


def test_selection_rules(lmax: int = 3):
    """Test Gaunt coefficient selection rules."""
    G_LLL = gaunt(lmax)
    L1max, L2max, L3max = G_LLL.shape

    for L1 in range(L1max):
        # Convert composit index to (l,m)
        l1, m1 = lm_indices(L1)
        for L2 in range(L2max):
            l2, m2 = lm_indices(L2)
            for L3 in range(L3max):
                l3, m3 = lm_indices(L3)
                # In order for the Gaunt coefficients to be finite, the
                # following conditions should be met
                if abs(G_LLL[L1, L2, L3]) > 1e-10:
                    assert (l1 + l2 + l3) % 2 == 0, \
                        f'l1 + l2 + l3 = {l1+l2+l3} should be an even integer'
                    assert abs(l1 - l2) <= l3 <= l1 + l2, f'{l1, l2, l3}'
                    assert m1 + m2 + m3 == 0 or \
                        m1 + m2 - m3 == 0 or \
                        m3 + m1 - m2 == 0 or \
                        m2 + m3 - m1 == 0, f'{m1, m2, m3}'


def test_permutation_symmetry(lmax: int = 3):
    """Test that the Gaunt coefficients are permutationally invariant"""
    G_LLL = gaunt(lmax)
    L1max, L2max, L3max = G_LLL.shape
    assert L2max == L3max

    for L1 in range(L1max):
        # Permutations between all indices
        for L2 in range(L1max):
            for L3 in range(L1max):
                for Lp1, Lp2, Lp3 in permutations([L1, L2, L3]):
                    assert abs(G_LLL[L1, L2, L3]
                               - G_LLL[Lp1, Lp2, Lp3]) < 1e-10
        # Permutations between L2 and L3
        for L2 in range(L2max):
            for L3 in range(L3max):
                assert abs(G_LLL[L1, L2, L3] - G_LLL[L1, L3, L2]) < 1e-10


def test_super_contraction_rule(lmax: int = 2):
    """Test that three spherical harmonics can be contracted to one."""
    G_LLLL = super_gaunt(lmax)
    L1max, L2max, L3max, L4max = G_LLLL.shape
    assert L1max == L2max

    x, y, z = unit_sphere_test_coordinates()
    for L1 in range(L1max):
        for L2 in range(L2max):
            for L3 in range(L3max):
                product = Y(L1, x, y, z) * Y(L2, x, y, z) * Y(L3, x, y, z)
                expansion = 0.
                for L4 in range(L4max):
                    expansion += G_LLLL[L1, L2, L3, L4] * Y(L4, x, y, z)
                assert expansion == pytest.approx(product)


def test_super_selection_rules(lmax: int = 2):
    """Test selection rules for products of Gaunt coefficients."""
    G_LLLL = super_gaunt(lmax)
    L1max, L2max, L3max, L4max = G_LLLL.shape

    for L1 in range(L1max):
        # Convert composit index to (l,m)
        l1, m1 = lm_indices(L1)
        for L2 in range(L2max):
            l2, m2 = lm_indices(L2)
            for L3 in range(L3max):
                l3, m3 = lm_indices(L3)
                for L4 in range(L4max):
                    l4, m4 = lm_indices(L4)
                    # In order for the Gaunt coefficients to be finite, the
                    # following conditions should be met
                    if abs(G_LLLL[L1, L2, L3, L4]) > 1e-10:
                        assert (l1 + l2 + l3 + l4) % 2 == 0, \
                            f'l1 + l2 + l3 + l4 = {l1+l2+l3+l4} ' \
                            'should be an even integer'
                        # The allowed l' range governs the allowed l4 range
                        lp_range = np.arange(abs(l1 - l2), l1 + l2 + 1)
                        assert np.min(abs(lp_range - l3)) <= l4 <= l1 + l2 + l3
                        # m' needs to be allowed and repeated
                        mpset1 = set([-m1 - m2, m1 + m2, -m1 + m2, m1 - m2])
                        mpset2 = set([-m3 - m4, m3 + m4, -m3 + m4, m3 - m4])
                        assert len(mpset1 | mpset2) < len(mpset1) + len(mpset2)


def test_super_permutation_symmetry(lmax: int = 2):
    """Test that the super Gaunt coefficients are permutationally invariant"""
    G_LLLL = super_gaunt(lmax)
    L1max, L2max, L3max, L4max = G_LLLL.shape
    assert L1max == L2max

    for L1 in range(L1max):
        for L2 in range(L1max):
            # Permutations between all indices
            for L3 in range(L1max):
                for L4 in range(L1max):
                    for Lp1, Lp2, Lp3, Lp4 in permutations([L1, L2, L3, L4]):
                        assert abs(G_LLLL[L1, L2, L3, L4]
                                   - G_LLLL[Lp1, Lp2, Lp3, Lp4]) < 1e-10
            # Permutations between L3 and L4
            for L3 in range(L3max):
                for L4 in range(L3max):
                    assert abs(G_LLLL[L1, L2, L3, L4]
                               - G_LLLL[L1, L2, L4, L3]) < 1e-10
        

def unit_sphere_test_coordinates():
    """Unit-sphere coordinates to test"""
    theta, phi = np.meshgrid(np.linspace(0, np.pi, 6),
                             np.linspace(0, 2 * np.pi, 11),
                             indexing='ij')
    x = np.sin(theta) * np.cos(phi)
    y = np.sin(theta) * np.sin(phi)
    z = np.cos(theta)
    return x, y, z


def lm_indices(L):
    l = int(np.sqrt(L))
    m = L - l * (l + 1)
    return l, m



import pytest
import numpy as np
from gpaw.response.g0w0 import G0W0


@pytest.mark.response
def test_ppa(in_tmp_dir, gpw_files, scalapack, needs_ase_master):
    ref_result = np.asarray([[[11.30094393, 21.62842077],
                              [5.33751513, 16.06905725],
                              [8.75269938, 22.46579489]]])
    gw = G0W0(gpw_files['bn_pw'],
              bands=(3, 5),
              nbands=9,
              nblocks=1,
              ecut=40,
              ppa=True)

    results = gw.calculate()
    np.testing.assert_allclose(results['qp'], ref_result, rtol=1e-03)



import pytest
from gpaw.mpi import world
from ase import Atoms
from gpaw import GPAW
from gpaw.utilities.sic import NSCFSIC

pytestmark = pytest.mark.skipif(world.size > 1,
                                reason='world.size > 1')


@pytest.mark.later
def test_sic_nscfsic(in_tmp_dir):
    atoms = ['He', 'Be']  # ,'Ne'] # Ne deviates already 2.5 eV
    EE = []
    EREF = [-79.4, -399.8, -3517.6]

    for a in atoms:
        s = Atoms(a)
        s.center(vacuum=4.0)
        calc = GPAW(mode='fd', h=0.15, txt=a + '.txt')
        s.calc = calc
        s.get_potential_energy()
        EE.append(NSCFSIC(calc).calculate())

    print("Difference to table VI of Phys. Rev. B 23, 5048 in eV")
    # https://journals.aps.org/prb/abstract/10.1103/PhysRevB.23.5048
    print("%10s%10s%10s%10s" % ("atom", "ref.", "gpaw", "diff"))
    for a, er, e in zip(atoms, EREF, EE):
        print("%10s%10.2f%10.2f%10.2f" % (a, er, e, er - e))
        assert abs(er - e) < 0.1
        # Arbitrary 0.1 eV tolerance for non-self consistent SIC
        # Note that Ne already deviates 2.5 eV



# Copyright (C) 2003  CAMP
# Copyright (C) 2014 R. Warmbier Materials for Energy Research Group,
# Wits University
# Please see the accompanying LICENSE file for further information.
from typing import Tuple

from ase.utils import gcd
import numpy as np

import _gpaw
import gpaw.mpi as mpi


def frac(f: float,
         n: int = 2 * 3 * 4 * 5,
         tol: float = 1e-6) -> Tuple[int, int]:
    """Convert to fraction.

    >>> frac(0.5)
    (1, 2)
    """
    if f == 0:
        return 0, 1
    x = n * f
    if abs(x - round(x)) > n * tol:
        raise ValueError
    x = int(round(x))
    d = gcd(x, n)
    return x // d, n // d


def sfrac(f: float) -> str:
    """Format as fraction.

    >>> sfrac(0.5)
    '1/2'
    >>> sfrac(2 / 3)
    '2/3'
    >>> sfrac(0)
    '0'
    """
    if f == 0:
        return '0'
    return '%d/%d' % frac(f)


class Symmetry:
    """Interface class for determination of symmetry, point and space groups.

    It also provides to apply symmetry operations to kpoint grids,
    wavefunctions and forces.
    """
    def __init__(self, id_a, cell_cv, pbc_c=np.ones(3, bool), tolerance=1e-7,
                 point_group=True, time_reversal=True, symmorphic=True,
                 allow_invert_aperiodic_axes=True):
        """Construct symmetry object.

        Parameters:

        id_a: list of int
            Numbered atomic types
        cell_cv: array(3,3), float
            Cartesian lattice vectors
        pbc_c: array(3), bool
            Periodic boundary conditions.
        tolerance: float
            Tolerance for symmetry determination.
        symmorphic: bool
            Switch for the use of non-symmorphic symmetries aka: symmetries
            with fractional translations.  Default is to use only symmorphic
            symmetries.
        point_group: bool
            Use point-group symmetries.
        time_reversal: bool
            Use time-reversal symmetry.
        tolerance: float
            Relative tolerance.

        Attributes:

        op_scc:
            Array of rotation matrices
        ft_sc:
            Array of fractional translation vectors
        a_sa:
            Array of atomic indices after symmetry operation
        has_inversion:
            (bool) Have inversion
        """

        self.id_a = id_a
        self.cell_cv = np.array(cell_cv, float)
        assert self.cell_cv.shape == (3, 3)
        self.pbc_c = np.array(pbc_c, bool)
        self.tol = tolerance
        self.symmorphic = symmorphic
        self.point_group = point_group
        self.time_reversal = time_reversal

        self.op_scc = np.identity(3, int).reshape((1, 3, 3))
        self.ft_sc = np.zeros((1, 3))
        self.a_sa = np.arange(len(id_a)).reshape((1, -1))
        self.has_inversion = False
        self.gcd_c = np.ones(3, int)

        # For reading old gpw-files:
        self.allow_invert_aperiodic_axes = allow_invert_aperiodic_axes

    def analyze(self, spos_ac):
        """Determine list of symmetry operations.

        First determine all symmetry operations of the cell. Then call
        ``prune_symmetries`` to remove those symmetries that are not satisfied
        by the atoms.

        It is not mandatory to call this method.  If not called, only
        time reversal symmetry may be used.
        """
        if self.point_group:
            self.find_lattice_symmetry()
            self.prune_symmetries_atoms(spos_ac)

    def find_lattice_symmetry(self):
        """Determine list of symmetry operations."""
        # Symmetry operations as matrices in 123 basis.
        # Operation is a 3x3 matrix, with possible elements -1, 0, 1, thus
        # there are 3**9 = 19683 possible matrices:
        combinations = 1 - np.indices([3] * 9)
        U_scc = combinations.reshape((3, 3, 3**9)).transpose((2, 0, 1))

        # The metric of the cell should be conserved after applying
        # the operation:
        metric_cc = self.cell_cv.dot(self.cell_cv.T)
        metric_scc = np.einsum('sij, jk, slk -> sil',
                               U_scc, metric_cc, U_scc,
                               optimize=True)
        mask_s = abs(metric_scc - metric_cc).sum(2).sum(1) <= self.tol
        U_scc = U_scc[mask_s]

        # Operation must not swap axes that don't have same PBC:
        pbc_cc = np.logical_xor.outer(self.pbc_c, self.pbc_c)
        mask_s = ~U_scc[:, pbc_cc].any(axis=1)
        U_scc = U_scc[mask_s]

        if not self.allow_invert_aperiodic_axes:
            # Operation must not invert axes that are not periodic:
            mask_s = (U_scc[:, np.diag(~self.pbc_c)] == 1).all(axis=1)
            U_scc = U_scc[mask_s]

        self.op_scc = U_scc
        self.ft_sc = np.zeros((len(self.op_scc), 3))

    def prune_symmetries_atoms(self, spos_ac):
        """Remove symmetries that are not satisfied by the atoms."""

        if len(spos_ac) == 0:
            self.a_sa = np.zeros((len(self.op_scc), 0), int)
            return

        # Build lists of atom numbers for each type of atom - one
        # list for each combination of atomic number, setup type,
        # magnetic moment and basis set:
        a_ij = {}
        for a, id in enumerate(self.id_a):
            if id in a_ij:
                a_ij[id].append(a)
            else:
                a_ij[id] = [a]

        a_j = a_ij[self.id_a[0]]  # just pick the first species

        # if supercell disable fractional translations:
        if not self.symmorphic:
            op_cc = np.identity(3, int)
            ftrans_sc = spos_ac[a_j[1:]] - spos_ac[a_j[0]]
            ftrans_sc -= np.rint(ftrans_sc)
            for ft_c in ftrans_sc:
                a_a = self.check_one_symmetry(spos_ac, op_cc, ft_c, a_ij)
                if a_a is not None:
                    self.symmorphic = True
                    break

        symmetries = []
        ftsymmetries = []

        # go through all possible symmetry operations
        for op_cc in self.op_scc:
            # first ignore fractional translations
            a_a = self.check_one_symmetry(spos_ac, op_cc, [0, 0, 0], a_ij)
            if a_a is not None:
                symmetries.append((op_cc, [0, 0, 0], a_a))
            elif not self.symmorphic:
                # check fractional translations
                sposrot_ac = np.dot(spos_ac, op_cc)
                ftrans_jc = sposrot_ac[a_j] - spos_ac[a_j[0]]
                ftrans_jc -= np.rint(ftrans_jc)
                for ft_c in ftrans_jc:
                    try:
                        nom_c, denom_c = np.array([frac(ft, tol=self.tol)
                                                   for ft in ft_c]).T
                    except ValueError:
                        continue
                    ft_c = nom_c / denom_c
                    a_a = self.check_one_symmetry(spos_ac, op_cc, ft_c, a_ij)
                    if a_a is not None:
                        ftsymmetries.append((op_cc, ft_c, a_a))
                        for c, d in enumerate(denom_c):
                            if self.gcd_c[c] % d != 0:
                                self.gcd_c[c] *= d

        # Add symmetry operations with fractional translations at the end:
        symmetries.extend(ftsymmetries)
        self.op_scc = np.array([sym[0] for sym in symmetries])
        self.ft_sc = np.array([sym[1] for sym in symmetries])
        self.a_sa = np.array([sym[2] for sym in symmetries])

        inv_cc = -np.eye(3, dtype=int)
        self.has_inversion = (self.op_scc == inv_cc).all(2).all(1).any()

    def check_one_symmetry(self, spos_ac, op_cc, ft_c, a_ij):
        """Checks whether atoms satisfy one given symmetry operation."""

        a_a = np.zeros(len(spos_ac), int)
        for a_j in a_ij.values():
            spos_jc = spos_ac[a_j]
            for a in a_j:
                spos_c = np.dot(spos_ac[a], op_cc)
                sdiff_jc = spos_c - spos_jc - ft_c
                sdiff_jc -= sdiff_jc.round()
                indices = np.where(abs(sdiff_jc).max(1) < self.tol)[0]
                if len(indices) == 1:
                    j = indices[0]
                    a_a[a] = a_j[j]
                else:
                    assert len(indices) == 0
                    return

        return a_a

    def check(self, spos_ac):
        """Check if positions satisfy symmetry operations."""

        nsymold = len(self.op_scc)
        self.prune_symmetries_atoms(spos_ac)
        if len(self.op_scc) < nsymold:
            raise RuntimeError('Broken symmetry!')

    def reduce(self, bzk_kc, comm=None):
        """Reduce k-points to irreducible part of the BZ.

        Returns the irreducible k-points and the weights and other stuff.

        """
        nbzkpts = len(bzk_kc)
        U_scc = self.op_scc
        nsym = len(U_scc)

        time_reversal = self.time_reversal and not self.has_inversion
        bz2bz_ks = map_k_points_fast(bzk_kc, U_scc, time_reversal,
                                     comm, self.tol)

        bz2bz_k = -np.ones(nbzkpts + 1, int)
        ibz2bz_k = []
        for k in range(nbzkpts - 1, -1, -1):
            # Reverse order looks more natural
            if bz2bz_k[k] == -1:
                bz2bz_k[bz2bz_ks[k]] = k
                ibz2bz_k.append(k)
        ibz2bz_k = np.array(ibz2bz_k[::-1])
        bz2bz_k = bz2bz_k[:-1].copy()

        bz2ibz_k = np.empty(nbzkpts, int)
        bz2ibz_k[ibz2bz_k] = np.arange(len(ibz2bz_k))
        bz2ibz_k = bz2ibz_k[bz2bz_k]

        weight_k = np.bincount(bz2ibz_k) * (1.0 / nbzkpts)

        # Symmetry operation mapping IBZ to BZ:
        sym_k = np.empty(nbzkpts, int)
        for k in range(nbzkpts):
            # We pick the first one found:
            try:
                sym_k[k] = np.where(bz2bz_ks[bz2bz_k[k]] == k)[0][0]
            except IndexError:
                print(nbzkpts)
                print(k)
                print(bz2bz_k)
                print(bz2bz_ks[bz2bz_k[k]])
                print(np.shape(np.where(bz2bz_ks[bz2bz_k[k]] == k)))
                print(bz2bz_k[k])
                print(bz2bz_ks[bz2bz_k[k]] == k)
                raise

        # Time-reversal symmetry used on top of the point group operation:
        if time_reversal:
            time_reversal_k = sym_k >= nsym
            sym_k %= nsym
        else:
            time_reversal_k = np.zeros(nbzkpts, bool)

        assert (ibz2bz_k[bz2ibz_k] == bz2bz_k).all()
        for k in range(nbzkpts):
            sign = 1 - 2 * time_reversal_k[k]
            dq_c = (np.dot(U_scc[sym_k[k]], bzk_kc[bz2bz_k[k]]) -
                    sign * bzk_kc[k])
            dq_c -= dq_c.round()
            assert abs(dq_c).max() < 1e-10

        return (bzk_kc[ibz2bz_k], weight_k,
                sym_k, time_reversal_k, bz2ibz_k, ibz2bz_k, bz2bz_ks)

    def check_grid(self, N_c) -> bool:
        """Check that symmetries are comensurate with grid."""
        for s, (U_cc, ft_c) in enumerate(zip(self.op_scc, self.ft_sc)):
            t_c = ft_c * N_c
            # Make sure all grid-points map onto another grid-point:
            if (((N_c * U_cc).T % N_c).any() or
                not np.allclose(t_c, t_c.round())):
                return False
        return True

    def symmetrize(self, a, gd):
        """Symmetrize array."""
        gd.symmetrize(a, self.op_scc, self.ft_sc)

    def symmetrize_positions(self, spos_ac):
        """Symmetrizes the atomic positions."""
        spos_tmp_ac = np.zeros_like(spos_ac)
        spos_new_ac = np.zeros_like(spos_ac)
        for i, op_cc in enumerate(self.op_scc):
            spos_tmp_ac[:] = 0.
            for a in range(len(spos_ac)):
                spos_c = np.dot(spos_ac[a], op_cc) - self.ft_sc[i]
                # Bring back the negative ones:
                spos_c = spos_c - np.floor(spos_c + 1e-5)
                spos_tmp_ac[self.a_sa[i][a]] += spos_c
            spos_new_ac += spos_tmp_ac

        spos_new_ac /= len(self.op_scc)
        return spos_new_ac

    def symmetrize_wavefunction(self, a_g, kibz_c, kbz_c, op_cc,
                                time_reversal):
        """Generate Bloch function from symmetry related function in the IBZ.

        a_g: ndarray
            Array with Bloch function from the irreducible BZ.
        kibz_c: ndarray
            Corresponding k-point coordinates.
        kbz_c: ndarray
            K-point coordinates of the symmetry related k-point.
        op_cc: ndarray
            Point group operation connecting the two k-points.
        time-reversal: bool
            Time-reversal symmetry required in addition to the point group
            symmetry to connect the two k-points.
        """

        # Identity
        if (np.abs(op_cc - np.eye(3, dtype=int)) < 1e-10).all():
            if time_reversal:
                return a_g.conj()
            else:
                return a_g
        # Inversion symmetry
        elif (np.abs(op_cc + np.eye(3, dtype=int)) < 1e-10).all():
            return a_g.conj()
        # General point group symmetry
        else:
            import _gpaw
            b_g = np.zeros_like(a_g)
            if time_reversal:
                # assert abs(np.dot(op_cc, kibz_c) - -kbz_c) < tol
                _gpaw.symmetrize_wavefunction(a_g, b_g, op_cc.T.copy(),
                                              kibz_c, -kbz_c)
                return b_g.conj()
            else:
                # assert abs(np.dot(op_cc, kibz_c) - kbz_c) < tol
                _gpaw.symmetrize_wavefunction(a_g, b_g, op_cc.T.copy(),
                                              kibz_c, kbz_c)
                return b_g

    def symmetrize_forces(self, F0_av):
        """Symmetrize forces."""
        F_ac = np.zeros_like(F0_av)
        for map_a, op_cc in zip(self.a_sa, self.op_scc):
            op_vv = np.dot(np.linalg.inv(self.cell_cv),
                           np.dot(op_cc, self.cell_cv))
            for a1, a2 in enumerate(map_a):
                F_ac[a2] += np.dot(F0_av[a1], op_vv)
        return F_ac / len(self.op_scc)

    def __str__(self):
        n = len(self.op_scc)
        nft = self.ft_sc.any(1).sum()
        lines = ['Symmetries present (total): {0}'.format(n)]
        if not self.symmorphic:
            lines.append(
                'Symmetries with fractional translations: {0}'.format(nft))

        # X-Y grid of symmetry matrices:

        lines.append('')
        nx = 6 if self.symmorphic else 3
        ns = len(self.op_scc)
        y = 0
        for y in range((ns + nx - 1) // nx):
            for c in range(3):
                line = ''
                for x in range(nx):
                    s = x + y * nx
                    if s == ns:
                        break
                    op_c = self.op_scc[s, c]
                    ft = self.ft_sc[s, c]
                    line += '  (%2d %2d %2d)' % tuple(op_c)
                    if not self.symmorphic:
                        line += ' + (%4s)' % sfrac(ft)
                lines.append(line)
            lines.append('')
        return '\n'.join(lines)


def map_k_points(bzk_kc, U_scc, time_reversal, comm=None, tol=1e-11):
    """Find symmetry relations between k-points.

    This is a Python-wrapper for a C-function that does the hard work
    which is distributed over comm.

    The map bz2bz_ks is returned.  If there is a k2 for which::

      = _    _    _
      U q  = q  + N,
       s k1   k2

    where N is a vector of integers, then bz2bz_ks[k1, s] = k2, otherwise
    if there is a k2 for which::

      = _     _    _
      U q  = -q  + N,
       s k1    k2

    then bz2bz_ks[k1, s + nsym] = k2, where nsym = len(U_scc).  Otherwise
    bz2bz_ks[k1, s] = -1.
    """

    if comm is None or isinstance(comm, mpi.DryRunCommunicator):
        comm = mpi.serial_comm

    nbzkpts = len(bzk_kc)
    ka = nbzkpts * comm.rank // comm.size
    kb = nbzkpts * (comm.rank + 1) // comm.size
    assert comm.sum(kb - ka) == nbzkpts

    if time_reversal:
        U_scc = np.concatenate([U_scc, -U_scc])

    bz2bz_ks = np.zeros((nbzkpts, len(U_scc)), int)
    bz2bz_ks[ka:kb] = -1
    _gpaw.map_k_points(np.ascontiguousarray(bzk_kc),
                       np.ascontiguousarray(U_scc), tol, bz2bz_ks, ka, kb)
    comm.sum(bz2bz_ks)
    return bz2bz_ks


def map_k_points_fast(bzk_kc, U_scc, time_reversal, comm=None, tol=1e-7):
    """Find symmetry relations between k-points.

    Performs the same task as map_k_points(), but much faster.
    This is achieved by finding the symmetry related kpoints using
    lexical sorting instead of brute force searching.

    bzk_kc: ndarray
        kpoint coordinates.
    U_scc: ndarray
        Symmetry operations
    time_reversal: Bool
        Use time reversal symmetry in mapping.
    comm:
        Communicator
    tol: float
        When kpoint are closer than tol, they are
        considered to be identical.
    """

    nbzkpts = len(bzk_kc)

    if time_reversal:
        U_scc = np.concatenate([U_scc, -U_scc])

    bz2bz_ks = np.zeros((nbzkpts, len(U_scc)), int)
    bz2bz_ks[:] = -1

    for s, U_cc in enumerate(U_scc):
        # Find mapped kpoints
        Ubzk_kc = np.dot(bzk_kc, U_cc.T)

        # Do some work on the input
        k_kc = np.concatenate([bzk_kc, Ubzk_kc])
        k_kc = np.mod(np.mod(k_kc, 1), 1)
        aglomerate_points(k_kc, tol)
        k_kc = k_kc.round(-np.log10(tol).astype(int))
        k_kc = np.mod(k_kc, 1)

        # Find the lexicographical order
        order = np.lexsort(k_kc.T)
        k_kc = k_kc[order]
        diff_kc = np.diff(k_kc, axis=0)
        equivalentpairs_k = np.array((diff_kc == 0).all(1),
                                     bool)

        # Mapping array.
        orders = np.array([order[:-1][equivalentpairs_k],
                           order[1:][equivalentpairs_k]])

        # This has to be true.
        assert (orders[0] < nbzkpts).all()
        assert (orders[1] >= nbzkpts).all()
        bz2bz_ks[orders[1] - nbzkpts, s] = orders[0]

    return bz2bz_ks


def aglomerate_points(k_kc, tol):
    nd = k_kc.shape[1]
    nbzkpts = len(k_kc)
    inds_kc = np.argsort(k_kc, axis=0)
    for c in range(nd):
        sk_k = k_kc[inds_kc[:, c], c]
        dk_k = np.diff(sk_k)

        # Partition the kpoints into groups
        pt_K = np.argwhere(dk_k > tol)[:, 0]
        pt_K = np.append(np.append(0, pt_K + 1), 2 * nbzkpts)
        for i in range(len(pt_K) - 1):
            k_kc[inds_kc[pt_K[i]:pt_K[i + 1], c],
                 c] = k_kc[inds_kc[pt_K[i], c], c]


def atoms2symmetry(atoms, id_a=None, tolerance=1e-7):
    """Create symmetry object from atoms object."""
    if id_a is None:
        id_a = atoms.get_atomic_numbers()
    symmetry = Symmetry(id_a, atoms.cell, atoms.pbc,
                        symmorphic=False,
                        time_reversal=False,
                        tolerance=tolerance)
    symmetry.analyze(atoms.get_scaled_positions())
    return symmetry


class CLICommand:
    """Analyse symmetry (and show IBZ k-points).

    Example:

        $ ase build -x bcc -a 3.5 Li | gpaw symmetry -k "{density:3,gamma:1}"
        symmetry:
          number of symmetries: 48
          number of symmetries with translation: 0

        bz sampling:
          number of bz points: 512
          number of ibz points: 29
          monkhorst-pack size: [8, 8, 8]
          monkhorst-pack shift: [0.0625, 0.0625, 0.0625]
    """

    @staticmethod
    def add_arguments(parser):
        parser.add_argument('-t', '--tolerance', type=float, default=1e-7,
                            help='Tolerance used for identifying symmetries.')
        parser.add_argument(
            '-k', '--k-points',
            help='Use symmetries to reduce number of k-points.  '
            'Exapmples: "4,4,4", "{density:3.5,gamma:True}".')
        parser.add_argument('-v', '--verbose', action='store_true',
                            help='Show symmetry operations (and k-points).')
        parser.add_argument('-s', '--symmorphic', action='store_true',
                            help='Only find symmorphic symmetries.')
        parser.add_argument('filename', nargs='?', default='-',
                            help='Filename to read structure from.  '
                            'Use "-" for reading from stdin.  '
                            'Default is "-".')

    @staticmethod
    def run(args):
        import sys
        from gpaw.new.symmetry import create_symmetries_object
        from gpaw.new.builder import create_kpts
        from gpaw.new.input_parameters import kpts
        from ase.cli.run import str2dict
        from ase.db import connect
        from ase.io import read

        if args.filename == '-':
            atoms = next(connect(sys.stdin).select()).toatoms()
        else:
            atoms = read(args.filename)
        symmetries = create_symmetries_object(
            atoms,
            parameters={'tolerance': args.tolerance,
                        'symmorphic': args.symmorphic})
        txt = str(symmetries)
        if not args.verbose:
            txt = txt.split('  rotations', 1)[0]
        print(txt)
        if args.k_points:
            k = str2dict('kpts=' + args.k_points)['kpts']
            bz = create_kpts(kpts(k), atoms)
            ibz = symmetries.reduce(bz)
            txt = str(ibz)
            if not args.verbose:
                txt = txt.split('  points and weights:', 1)[0]
            print(txt)



import numpy as np
from gpaw.core import PWDesc
from gpaw.gpu import cupy as cp
from gpaw.mpi import broadcast_float
from gpaw.new import zips, spinsum
from gpaw.new.pot_calc import PotentialCalculator
from gpaw.new.pw.stress import calculate_stress
from gpaw.setup import Setups
from gpaw.new.calculation import DFTState


class PlaneWavePotentialCalculator(PotentialCalculator):
    def __init__(self,
                 grid,
                 fine_grid,
                 pw: PWDesc,
                 fine_pw: PWDesc,
                 setups: Setups,
                 xc,
                 poisson_solver,
                 *,
                 fracpos_ac,
                 atomdist,
                 soc=False,
                 xp=np):
        self.xp = xp
        super().__init__(xc, poisson_solver, setups,
                         fracpos_ac=fracpos_ac,
                         soc=soc)

        self.vbar_ag = setups.create_local_potentials(
            pw, fracpos_ac, atomdist, xp)
        self.ghat_aLh = setups.create_compensation_charges(
            fine_pw, fracpos_ac, atomdist, xp)

        self.pw = pw
        self.fine_pw = fine_pw
        self.pw0 = pw.new(comm=None)  # not distributed

        self.h_g, self.g_r = fine_pw.map_indices(self.pw0)
        if xp is cp:
            self.h_g = cp.asarray(self.h_g)
            self.g_r = [cp.asarray(g) for g in self.g_r]

        self.fftplan = grid.fft_plans(xp=xp)
        self.fftplan2 = fine_grid.fft_plans(xp=xp)

        self.grid = grid
        self.fine_grid = fine_grid

        self.vbar_g = pw.zeros(xp=xp)
        self.vbar_ag.add_to(self.vbar_g)
        self.vbar0_g = self.vbar_g.gather()

        # For forces and stress:
        self._nt_g = None
        self._vt_g = None
        self._dedtaut_g = None

    def interpolate(self, a_R, a_r=None):
        return a_R.interpolate(self.fftplan, self.fftplan2,
                               grid=self.fine_grid, out=a_r)

    def restrict(self, a_r, a_R=None):
        return a_r.fft_restrict(self.fftplan2, self.fftplan,
                                grid=self.grid, out=a_R)

    def calculate_charges(self, vHt_h):
        return self.ghat_aLh.integrate(vHt_h)

    def _interpolate_density(self, nt_sR):
        nt_sr = self.fine_grid.empty(nt_sR.dims, xp=self.xp)
        pw = self.vbar_g.desc

        if pw.comm.rank == 0:
            indices = self.xp.asarray(self.pw0.indices(self.fftplan.shape))
            nt0_g = self.pw0.zeros(xp=self.xp)
        else:
            nt0_g = None

        ndensities = nt_sR.dims[0] % 3
        for spin, (nt_R, nt_r) in enumerate(zips(nt_sR, nt_sr)):
            self.interpolate(nt_R, nt_r)
            if spin < ndensities and pw.comm.rank == 0:
                nt0_g.data += self.xp.asarray(
                    self.fftplan.tmp_Q.ravel()[indices])

        return nt_sr, pw, nt0_g

    def _interpolate_and_calculate_xc(self, xc, nt_sR, ibzwfs):
        ...

    def calculate_non_selfconsistent_exc(self, xc, nt_sR, ibzwfs):
        _, _, _, _, e_xc, _ = self._interpolate_and_calculate_xc(
            xc, nt_sR, ibzwfs)
        return e_xc

    def calculate_pseudo_potential(self, density, ibzwfs, vHt_h=None):
        nt_sr, pw, nt0_g = self._interpolate_density(density.nt_sR)

        if density.taut_sR is not None:
            taut_sr = self.interpolate(density.taut_sR)
        else:
            taut_sr = None

        e_xc, vxct_sr, dedtaut_sr = self.xc.calculate(nt_sr, taut_sr)

        if pw.comm.rank == 0:
            nt0_g.data *= 1 / np.prod(density.nt_sR.desc.size_c)
            e_zero = self.vbar0_g.integrate(nt0_g)
        else:
            e_zero = 0.0
        e_zero = broadcast_float(float(e_zero), pw.comm)

        if vHt_h is None:
            vHt_h = self.ghat_aLh.pw.zeros(xp=self.xp)

        charge_h = vHt_h.desc.zeros(xp=self.xp)
        coef_aL = density.calculate_compensation_charge_coefficients()
        self.ghat_aLh.add_to(charge_h, coef_aL)

        if pw.comm.rank == 0:
            for rank, g in enumerate(self.g_r):
                if rank == 0:
                    charge_h.data[self.h_g] += nt0_g.data[g]
                else:
                    pw.comm.send(nt0_g.data[g], rank)
        else:
            data = self.xp.empty(len(self.h_g), complex)
            pw.comm.receive(data, 0)
            charge_h.data[self.h_g] += data

        # background charge ???

        e_coulomb = self.poisson_solver.solve(vHt_h, charge_h)

        if pw.comm.rank == 0:
            vt0_g = self.vbar0_g.copy()
            for rank, g in enumerate(self.g_r):
                if rank == 0:
                    vt0_g.data[g] += vHt_h.data[self.h_g]
                else:
                    data = self.xp.empty(len(g), complex)
                    pw.comm.receive(data, rank)
                    vt0_g.data[g] += data
            vt0_R = vt0_g.ifft(
                plan=self.fftplan,
                grid=density.nt_sR.desc.new(comm=None))
        else:
            pw.comm.send(vHt_h.data[self.h_g], 0)

        vt_sR = density.nt_sR.new()
        vt_sR[0].scatter_from(vt0_R if pw.comm.rank == 0 else None)
        if density.ndensities == 2:
            vt_sR.data[1] = vt_sR.data[0]
        vt_sR.data[density.ndensities:] = 0.0

        vtmp_R = vt_sR.desc.empty(xp=self.xp)
        for spin, (vt_R, vxct_r) in enumerate(zips(vt_sR, vxct_sr)):
            self.restrict(vxct_r, vtmp_R)
            vt_R.data += vtmp_R.data

        e_external = 0.0

        self._reset()

        return {'coulomb': e_coulomb,
                'zero': e_zero,
                'xc': e_xc,
                'stress': e_coulomb + e_zero,
                'external': e_external}, vt_sR, dedtaut_sr, vHt_h

    def move(self, fracpos_ac, atomdist):
        self.ghat_aLh.move(fracpos_ac, atomdist)
        self.vbar_ag.move(fracpos_ac, atomdist)
        self.vbar_g.data[:] = 0.0
        self.vbar_ag.add_to(self.vbar_g)
        self.vbar0_g = self.vbar_g.gather()
        self._reset()

    def _reset(self):
        self._vt_g = None
        self._nt_g = None
        self._dedtaut_g = None

    def _force_stress_helper(self, state: DFTState):
        """Only do the work once - in case both forces and stress is needed"""
        if self._vt_g is not None:
            return self._vt_g, self._nt_g, self._dedtaut_g

        density = state.density
        potential = state.potential
        nt_R = spinsum(density.nt_sR)
        vt_R = spinsum(potential.vt_sR, mean=True)
        self._vt_g = vt_R.fft(self.fftplan, pw=self.pw)
        self._nt_g = nt_R.fft(self.fftplan, pw=self.pw)

        dedtaut_sR = potential.dedtaut_sR
        if dedtaut_sR is not None:
            dedtaut_R = spinsum(dedtaut_sR, mean=True)
            self._dedtaut_g = dedtaut_R.fft(self.fftplan, pw=self.pw)
        else:
            self._dedtaut_g = None

        return self._vt_g, self._nt_g, self._dedtaut_g

    def force_contributions(self, state):
        vt_g, nt_g, dedtaut_g = self._force_stress_helper(state)
        if dedtaut_g is None:
            Ftauct_av = None
        else:
            Ftauct_av = state.density.tauct_aX.derivative(dedtaut_g)

        return (self.ghat_aLh.derivative(state.potential.vHt_x),
                state.density.nct_aX.derivative(vt_g),
                Ftauct_av,
                self.vbar_ag.derivative(nt_g))

    def stress(self, state):
        vt_g, nt_g, dedtaut_g = self._force_stress_helper(state)
        return calculate_stress(self, state, vt_g, nt_g, dedtaut_g)



# Copyright (C) 2008  CAMd
# Please see the accompanying LICENSE file for further information.

"""This module is an example of good coding style.

This docstring should begin with a one-line description followed by a
blank line, and then this paragraph describing in more word what kind
of functionality this module implements.

After this docstring we have import statements in this order:

1. From the Python standard library.
2. Other libraries (numpy, ase, ...).
3. GPAW stuff.
"""

from math import pi

import numpy as np
from ase.units import Ha

from gpaw import debug
import gpaw.mpi as mpi


class SimpleExample:
    """A simple example class.

    A headline, a blank line and then this longer description of the
    class.

    Here one could put an example of how to use the class::

      ex = SimpleExample('Test', (2, 3), int, verbose=False)
      ex.run(7, verbose=True)

    """

    def __init__(self, name, shape, dtype=float, verbose=True):
        """Create an example object.

        Again, headline, blank line, ... .  If there are many
        parameters, there should be a parameter section (see below).
        If there only a few possible arguments, then the parameter
        section can be left out and the arguments can be described in
        the section following the headline and blank line (see the
        `run` method).  If a method is real simple and
        self-explanatory, the docstring can be the headline only (see
        the `reset` method).

        Parameters:

        name : string
            Name of the example.
        shape:  tuple
            Shape of the ndarray.
        dtype: ndarray datatype
            The datatype of the ndarray.  Here, the description can go
            on to a second line if needed.  Make sure that the
            indentation is like shown here, and remember to end with a
            period.
        verbose: boolean
            Print information about this and that.

        Other sections:

        There can be other sections - see bolow and here:

          http://scipy.org/...

        """

        self.name = name
        if verbose:
            print(name)
        self.a = np.zeros(shape, dtype)
        self.verbose = verbose

    def method_with_long_name(self, b, out=None):
        """Do something very complicated.

        Long story with all details here ...

        Parameters:

        b: ndarray
            Add this array.
        out: ndarray
            Optional output array.

        Returns:

        The sum of ...
        """

        if out is None:
            return self.a + b
        else:
            return np.add(self.a, b, out)

    def run(self, n):
        """Do something.

        Do it n times, where n must be a positive integer.  The final
        result bla-bla is returned.
        """

        for i in range(n):
            self.a += i
            if self.verbose:
                print(self.a)

        return pi * self.a / n + 1

    def reset(self):
        """Simple method - no explanation needed."""
        self.a[:] = 0


def function(a, b):
    """Headline.

    Long story ..."""

    result = a + b
    if debug and mpi.world.rank == 0:
        print(result * Ha)
    return result



"""Edmiston-ruedenberg localization."""
from math import pi

import numpy as np

import _gpaw
from .overlaps import WannierOverlaps
from .functions import WannierFunctions


class LocalizationNotConvergedError(Exception):
    """Error raised if maxiter is exceeded."""


def localize(overlaps: WannierOverlaps,
             maxiter: int = 100,
             tolerance: float = 1e-5,
             verbose: bool = not False) -> WannierFunctions:
    """Simple localization.

    Orthorhombic cell, no k-points, only occupied states.
    """
    if not overlaps.atoms.cell.orthorhombic:
        raise NotImplementedError('An orthogonal cell is required')
    assert (overlaps.monkhorst_pack_size == (1, 1, 1)).all()

    Z_nnc = np.empty((overlaps.nbands, overlaps.nbands, 3), complex)
    for c, direction in enumerate([(1, 0, 0), (0, 1, 0), (0, 0, 1)]):
        Z_nnc[:, :, c] = overlaps.overlap(bz_index=0, direction=direction)

    U_nn = np.identity(overlaps.nbands)

    if verbose:
        print('iter      value     change')
        print('---- ---------- ----------')

    old = 0.0
    for iter in range(maxiter):
        value = _gpaw.localize(Z_nnc, U_nn)
        if verbose:
            print(f'{iter:4} {value:10.3f} {value - old:10.6f}')
        if value - old < tolerance:
            break
        old = value
    else:
        raise LocalizationNotConvergedError(
            f'Did not converge in {maxiter} iterations')

    # Find centers:
    scaled_nc = -np.angle(Z_nnc.diagonal()).T / (2 * pi)
    centers_nv = (scaled_nc % 1.0).dot(overlaps.atoms.cell)

    return WannierFunctions(overlaps.atoms, centers_nv, value, [U_nn])


if __name__ == '__main__':
    import sys
    from gpaw import GPAW
    from .overlaps import calculate_overlaps

    calc = GPAW(sys.argv[1])
    nwannier = int(sys.argv[2])
    overlaps = calculate_overlaps(calc, nwannier)
    wan = overlaps.localize_er(verbose=True)
    print(wan.centers)
    wan.centers_as_atoms().edit()



"""Parallelization scheme for frequency–planewave–planewave arrays."""
from gpaw.mpi import world
from gpaw.response.pw_parallelization import block_partition
from gpaw.utilities.scalapack import scalapack_set, scalapack_solve
from gpaw.blacs import BlacsGrid
import numpy as np


def get_blocksize(length, commsize):
    return -(-length // commsize)


def get_strides(cpugrid):
    return np.array([cpugrid[1] * cpugrid[2], cpugrid[2], 1], int)


class Grid:
    def __init__(self, comm, shape, cpugrid=None, blocksize=None):
        self.comm = comm
        self.shape = shape

        if cpugrid is None:
            cpugrid = choose_parallelization(shape[0], shape[1],
                                             comm.size)

        self.cpugrid = cpugrid

        if blocksize is None:
            blocksize = [get_blocksize(size, commsize)
                         for size, commsize in zip(shape, cpugrid)]
            # XXX scalapack blocksize hack
            blocksize[1] = blocksize[2] = max(blocksize[1:])

        # XXX Our scalapack interface does NOT like it when blocksizes
        # are not the same.  There must be a bug.
        assert blocksize[1] == blocksize[2]

        self.blocksize = tuple(blocksize)
        self.cpugrid = cpugrid

        self.myparpos = self.rank2parpos(self.comm.rank)

        n_cp = self.get_domains()

        shape = []
        myslice_c = []
        for i in range(3):
            n_p = n_cp[i]
            parpos = self.myparpos[i]
            myend = n_p[parpos + 1]
            mystart = n_p[parpos]
            myslice_c.append(slice(mystart, myend))
            size = myend - mystart
            shape.append(size)

        self.myshape = tuple(shape)
        self.myslice = tuple(myslice_c)

    # TODO inherit these from array descriptor
    def zeros(self, dtype=float):
        return np.zeros(self.myshape, dtype=dtype)

    def get_domains(self):
        """Get definition of domains.

        Returns domains_cp where domains_cp[c][r + 1] - domains_cp[c][r]
        is the number of points in domain r along direction c.

        The second axis contains the "fencepost" locations
        of the grid: [0, blocksize, 2 * blocksize, ...]
        """
        domains_cp = []

        for i in range(3):
            n_p = np.empty(self.cpugrid[i] + 1, int)
            n_p[0] = 0
            n_p[1:] = self.blocksize[i]
            n_p[:] = n_p.cumsum().clip(0, self.shape[i])
            domains_cp.append(n_p)

        return domains_cp

    def rank2parpos(self, rank):
        # XXX Borrowing from gd -- we should eliminate this duplication.

        strides = get_strides(self.cpugrid)
        cpugrid_coord = np.array(
            [rank // strides[0],
             (rank % strides[0]) // strides[1],
             rank % strides[1]])

        return cpugrid_coord

    def redistribute(self, dstgrid, srcarray, dstarray):
        from gpaw.utilities.grid_redistribute import general_redistribute
        domains1 = self.get_domains()
        domains2 = dstgrid.get_domains()
        general_redistribute(self.comm, domains1, domains2,
                             self.rank2parpos, dstgrid.rank2parpos,
                             srcarray, dstarray, behavior='overwrite')

    def invert_inplace(self, x_wgg):
        # Build wgg grid choosing scalapack
        nscalapack_cores = np.prod(self.cpugrid[1:])
        blacs_comm, wcomm = block_partition(self.comm, nscalapack_cores)
        assert wcomm.size == self.cpugrid[0]
        assert blacs_comm.size * wcomm.size == self.comm.size
        for iw, x_gg in enumerate(x_wgg):
            bg = BlacsGrid(blacs_comm, *self.cpugrid[1:][::-1])
            desc = bg.new_descriptor(
                *self.shape[1:],
                *self.blocksize[1:])

            xtmp_gg = desc.empty(dtype=x_wgg.dtype)
            xtmp_gg[:] = x_gg.T

            righthand = desc.zeros(dtype=complex)
            scalapack_set(desc, righthand, alpha=0.0, beta=1.0, uplo='U')

            scalapack_solve(desc, desc, xtmp_gg, righthand)
            x_gg[:] = righthand.T


def get_x_WGG(WGG_grid):
    x_WGG = WGG_grid.zeros(dtype=complex)
    rng = np.random.RandomState(42)

    x_WGG.flat[:] = rng.random(x_WGG.size)
    x_WGG.flat[:] += rng.random(x_WGG.size) * 1j
    # XXX write also to imaginary parts

    nG = x_WGG.shape[1]

    xinv_WGG = np.zeros_like(x_WGG)
    if WGG_grid.comm.rank == 0:
        assert x_WGG.shape == WGG_grid.myshape
        for iw, x_GG in enumerate(x_WGG):
            x_GG += x_GG.T.conj().copy()
            x_GG += np.identity(nG) * 5
            eigs = np.linalg.eigvals(x_GG)
            assert all(eigs.real) > 0
            xinv_WGG[iw] = np.linalg.inv(x_GG)
    else:
        assert np.prod(x_WGG.shape) == 0
    return x_WGG, xinv_WGG


def factorize(N):
    for n in range(1, N + 1):
        if N % n == 0:
            yield N // n, n


def get_products(N):
    for a1, a2 in factorize(N):
        for a2p, a3 in factorize(a2):
            yield a1, a2p, a3


def choose_parallelization(nW, nG, commsize):
    min_badness = 10000000

    for wGG in get_products(commsize):
        wsize, gsize1, gsize2 = wGG
        nw = (nW + wsize - 1) // wsize

        if nw > nW:
            continue

        number_of_cores_with_zeros = (wsize * nw - nW) // nw
        scalapack_skew = (gsize1 - gsize2)**2
        scalapack_size = gsize1 * gsize2
        badness = (number_of_cores_with_zeros * 1000
                   + 10 * scalapack_skew + scalapack_size)

        # print(wsize, gsize1, gsize2, nw, number_of_cores_with_zeros, badness)
        if badness < min_badness:
            wGG_min = wGG
            min_badness = badness
    return wGG_min


def main(comm=world):
    nW = 3
    nG = 31

    cpugrid = choose_parallelization(nW, nG, comm.size)

    WGG = (nW, nG, nG)
    dtype = complex

    # Build serial grid (data only on rank 0)
    # and establish matrix and its inverse
    WGG_grid = Grid(comm, WGG, cpugrid, blocksize=WGG)
    x_WGG, xinv_WGG = get_x_WGG(WGG_grid)

    # Distribute to WgG grid:
    WgG_grid = Grid(comm, WGG, (1, comm.size, 1))
    x_WgG = np.zeros(WgG_grid.myshape, dtype=dtype)
    WGG_grid.redistribute(WgG_grid, x_WGG, x_WgG)

    wgg_grid = Grid(comm, WGG, cpugrid)
    print(f'cpugrid={cpugrid} blocksize={wgg_grid.blocksize} '
          f'shape={wgg_grid.shape} myshape={wgg_grid.myshape}')

    x_wgg = wgg_grid.zeros(dtype=dtype)
    WgG_grid.redistribute(wgg_grid, x_WgG, x_wgg)

    # By now let's distribute wgg back to WgG to check that numbers
    # are the same:
    x1_WgG = WgG_grid.zeros(dtype=dtype)
    wgg_grid.redistribute(WgG_grid, x_wgg, x1_WgG)
    assert np.allclose(x_WgG, x1_WgG)

    wgg_grid.invert_inplace(x_wgg)

    # Distribute the inverse wgg back to WGG:
    inv_x_WGG = WGG_grid.zeros(dtype=dtype)
    wgg_grid.redistribute(WGG_grid, x_wgg, inv_x_WGG)

    from gpaw.utilities.tools import tri2full
    if comm.rank == 0:
        for inv_x_GG in inv_x_WGG:
            tri2full(inv_x_GG, 'L')

        for x_GG, inv_x_GG in zip(x_WGG, inv_x_WGG):
            assert np.allclose(x_GG @ inv_x_GG, np.identity(nG))


if __name__ == '__main__':
    main()



"""Calculate dipole matrix elements."""
from __future__ import annotations
import numpy as np
from ase.units import Bohr
from gpaw.new.ase_interface import ASECalculator, GPAW
from gpaw.typing import Array3D, Vector


def dipole_matrix_elements(*args, **kwargs):
    """Deprecated.

    Use
    ``gpaw.new.pwfd.wave_functions.PWFDWaveFunctions.dipole_matrix_elements``
    instead.
    """
    raise DeprecationWarning


def dipole_matrix_elements_from_calc(calc: ASECalculator,
                                     n1: int,
                                     n2: int,
                                     center: Vector = None
                                     ) -> list[Array3D]:
    """Calculate dipole matrix-elements (units: eÅ).

    Parameters
    ----------
    n1, n2:
        Band range.
    center:
        Center of molecule in Å.  Defaults to center of cell.
    """
    ibzwfs = calc.calculation.state.ibzwfs

    assert ibzwfs.ibz.bz.gamma_only

    if center is not None:
        center = np.asarray(center) / Bohr

    wfs_s = ibzwfs.wfs_qs[0]

    d_snnv = []
    for wfs in wfs_s:
        if calc.params.mode['name'] == 'lcao':
            basis = calc.calculation.scf_loop.hamiltonian.basis
            grid = calc.calculation.state.density.nt_sR.desc
            wfs = wfs.to_uniform_grid_wave_functions(grid, basis)
        wfs = wfs.collect(n1, n2)
        if wfs is not None:
            d_nnv = wfs.dipole_matrix_elements(center) * Bohr
        else:
            d_nnv = np.empty((n2 - n1, n2 - n1, 3))
        calc.comm.broadcast(d_nnv, 0)
        d_snnv.append(d_nnv)

    return d_snnv


def main(argv: list[str] = None) -> None:
    import argparse

    parser = argparse.ArgumentParser(
        prog='python3 -m gpaw.utilities.dipole',
        description='Calculate dipole matrix elements.  Units: eÅ.')

    add = parser.add_argument

    add('file', metavar='input-file',
        help='GPW-file with wave functions.')
    add('-n', '--band-range', nargs=2, type=int, default=[0, 0],
        metavar=('n1', 'n2'), help='Include bands: n1 <= n < n2.')
    add('-c', '--center', metavar='x,y,z',
        help='Center of charge distribution.  Default is middle of unit '
        'cell.')

    args = parser.parse_intermixed_args(argv)

    calc = GPAW(args.file)

    n1, n2 = args.band_range
    nbands = calc.get_number_of_bands()
    n2 = n2 or n2 + nbands

    if args.center:
        center = [float(x) for x in args.center.split(',')]
    else:
        center = None  # center of cell

        d_snnv = dipole_matrix_elements_from_calc(calc, n1, n2, center)

    if calc.comm.rank > 0:
        return

    print('Number of bands:', nbands)
    print('Number of valence electrons:', calc.get_number_of_electrons())
    print('Units: eÅ')
    print()

    for spin, d_nnv in enumerate(d_snnv):
        print(f'Spin={spin}')

        for direction, d_nn in zip('xyz', d_nnv.T):
            print(f' <{direction}>',
                  ''.join(f'{n:8}' for n in range(n1, n2)))
            for n in range(n1, n2):
                print(f'{n:4}', ''.join(f'{d:8.3f}' for d in d_nn[n - n1]))


if __name__ == '__main__':
    main()



import pytest
from ase import Atom, Atoms

from gpaw import GPAW, restart, Davidson, Mixer
from gpaw.atom.generator import Generator
from gpaw.atom.configurations import parameters
from gpaw.mpi import world
from gpaw.test import equal

# This test calculates the derivative discontinuity of Ne-atom
# first on 3D without restart. Then does restart and recalculates.


@pytest.mark.gllb
@pytest.mark.libxc
def test_gllb_ne_disc(in_tmp_dir, add_cwd_to_setup_paths):
    atom = 'Ne'

    for xcname in ['GLLB', 'GLLBSC']:
        if world.rank == 0:
            g = Generator(atom, xcname=xcname, scalarrel=False, nofiles=True)
            g.run(**parameters[atom])
            eps = g.e_j[-1]
        world.barrier()

        a = 10
        Ne = Atoms([Atom(atom, (0, 0, 0))],
                   cell=(a, a, a), pbc=False)
        Ne.center()
        calc = GPAW(mode='fd',
                    eigensolver=Davidson(4),
                    nbands=10,
                    h=0.18,
                    xc=xcname,
                    basis='dzp',
                    mixer=Mixer(0.6))
        Ne.calc = calc
        Ne.get_potential_energy()
        homo, lumo = calc.get_homo_lumo()
        response = calc.hamiltonian.xc.response
        dxc_pot = response.calculate_discontinuity_potential(homo, lumo)
        KS, dxc = response.calculate_discontinuity(dxc_pot)
        if xcname == 'GLLB':
            equal(KS + dxc, 24.89, 1.5e-1)
        else:
            equal(KS + dxc, 27.70, 6.0e-2)
        eps3d = calc.wfs.kpt_u[0].eps_n[3]
        calc.write('Ne_temp.gpw', mode='all')

        atoms, calc = restart('Ne_temp.gpw')
        homo, lumo = calc.get_homo_lumo()
        response = calc.hamiltonian.xc.response
        dxc_pot = response.calculate_discontinuity_potential(homo, lumo)
        KS2, dxc2 = response.calculate_discontinuity(dxc_pot)
        equal(KS, KS2, 1e-5)
        equal(dxc2, dxc, 1e-5)

        # Hardness of Ne 24.71eV by GLLB+Dxc, experimental I-A = I = 21.56eV
        #
        # Not sure where 24.71 comes from, but with better grid and better
        # stencil, result becomes 24.89.  --askhl

        if world.rank == 0:
            equal(eps, eps3d, 1e-3)
        if xcname == 'GLLB':
            equal(24.89, KS2 + dxc2, 1.2e-1)



"""CLI-code for dos-subcommand."""
from pathlib import Path
from typing import Union, List, Tuple, Optional

import numpy as np
from ase.spectrum.dosdata import GridDOSData
from ase.spectrum.doscollection import GridDOSCollection

from gpaw import GPAW
from gpaw.setup import Setup
from gpaw.spherical_harmonics import names as ylmnames
from gpaw.dos import DOSCalculator


class CLICommand:
    """Calculate (projected) density of states from gpw-file."""

    @staticmethod
    def add_arguments(parser):
        add = parser.add_argument
        add('gpw', metavar='gpw-file')
        add('-p', '--plot', action='store_true',
            help='Plot the DOS.')
        add('-i', '--integrated', action='store_true',
            help='Calculate integrated DOS.')
        add('-w', '--width', type=float, default=0.1,
            help='Width of Gaussian.  Use 0.0 to use linear tetrahedron '
            'interpolation.')
        add('-a', '--atom', help='Project onto atoms: "Cu-spd,H-s" or use '
            'atom indices "12-spdf".  Particular m-values can be obtained '
            'like this: "N-p0,N-p1,N-p2. For p-orbitals, m=0,1,2 translates '
            'to y, z and x. For d-orbitals, m=0,1,2,3,4 translates '
            'to xy, yz, 3z2-r2, zx and x2-y2.')
        add('-t', '--total', action='store_true',
            help='Show both PDOS and total DOS.')
        add('-r', '--range', nargs=2, metavar=('emin', 'emax'),
            help='Energy range in eV relative to Fermi level.')
        add('-n', '--points', type=int, default=400, help='Number of points.')
        add('--soc', action='store_true',
            help='Include spin-orbit coupling.')

    @staticmethod
    def run(args):
        if args.range is None:
            emin = None
            emax = None
        else:
            emin, emax = (float(e) for e in args.range)
        dos(args.gpw,
            plot=args.plot,
            width=args.width,
            integrated=args.integrated,
            projection=args.atom,
            soc=args.soc,
            emin=emin,
            emax=emax,
            npoints=args.points,
            show_total=args.total)


def parse_projection_string(projection: str,
                            symbols: List[str],
                            setups: List[Setup]
                            ) -> List[Tuple[str,
                                            List[Tuple[int,
                                                       int,
                                                       Union[None, int]]]]]:
    """Create labels and lists of (a, l, m)-tuples.

    Example:

    >>> from gpaw.setup import create_setup
    >>> setup = create_setup('Li')
    >>> s, py = parse_projection_string('Li-sp0',
    ...                                 ['Li', 'Li'],
    ...                                 [setup, setup])
    >>> s
    ('Li-s', [(0, 0, None), (1, 0, None)])
    >>> py
    ('Li-p(y)', [(0, 1, 0), (1, 1, 0)])

    * "Li-s" will have contributions from l=0 and atoms 0 and 1
    * "Li-p(y)" will have contributions from l=1, m=0 and atoms 0 and 1

    """
    result: List[Tuple[str, List[Tuple[int, int, Union[None, int]]]]] = []
    for proj in projection.split(','):
        s, ll = proj.split('-')
        if s.isdigit():
            A = [int(s)]
            s = '#' + s
        else:
            A = [a for a, symbol in
                 enumerate(symbols)
                 if symbol == s]
            if not A:
                raise ValueError('No such atom: ' + s)
        for l, m in parse_lm_string(ll):
            label = s + '-' + 'spdfg'[l]
            if m is not None:
                name = ylmnames[l][m]
                label += f'({name})'
            result.append((label, [(a, l, m) for a in A]))

    return result


def parse_lm_string(s: str) -> List[Tuple[int, Union[None, int]]]:
    """Parse 'spdf' kind of string to numbers.

    Return list of (l, m) tuples with m=None if not specified:

    >>> parse_lm_string('sp')
    [(0, None), (1, None)]
    >>> parse_lm_string('p0p1p2')
    [(1, 0), (1, 1), (1, 2)]
    """
    result = []
    while s:
        l = 'spdfg'.index(s[0])
        m: Union[None, int]
        if s[1:2].isnumeric():
            m = int(s[1:2])
            s = s[2:]
        else:
            m = None
            s = s[1:]
        result.append((l, m))
    return result


def dos(filename: Union[Path, str],
        *,
        plot=False,
        width=0.1,
        integrated=False,
        projection=None,
        soc=False,
        emin=None,
        emax=None,
        npoints=200,
        show_total=None):
    """Calculate density of states.

    filename: str
        Name of restart-file.
    plot: bool
        Show a plot.
    width: float
        Width of Gaussians.
    integrated: bool
        Calculate integrated DOS.

    """
    calc = GPAW(filename)

    doscalc = DOSCalculator.from_calculator(calc, soc)
    energies = doscalc.get_energies(emin, emax, npoints)
    nspins = doscalc.nspins
    spinlabels = [''] if nspins == 1 else [' up', ' dn']
    spins: List[Optional[int]] = [None] if nspins == 1 else [0, 1]

    dosobjs = GridDOSCollection([], energies)

    # Note: ignoring wrong GridDOSCollection.__add__ hint below.

    if projection is None or show_total:
        for spin, label in zip(spins, spinlabels):
            dos = doscalc.raw_dos(energies, spin=spin, width=width)
            dosobjs += GridDOSData(energies, dos,  # type: ignore
                                   {'label': 'DOS' + label})

    if projection is not None:
        symbols = calc.atoms.get_chemical_symbols()
        projs = parse_projection_string(
            projection, symbols, calc.setups)
        for label, contributions in projs:
            for spin, spinlabel in zip(spins, spinlabels):
                dos = np.zeros_like(energies)
                for a, l, m in contributions:
                    dos += doscalc.raw_pdos(energies,
                                            a, l, m,
                                            spin=spin,
                                            width=width)
                dosobjs += GridDOSData(energies, dos,  # type: ignore
                                       {'label': label + spinlabel})

    if integrated:
        de = energies[1] - energies[0]
        energies = energies + de / 2
        dosobjs = GridDOSCollection(
            [GridDOSData(energies,
                         np.cumsum(obj.get_weights()) * de,
                         obj.info)
             for obj in dosobjs])
        ylabel = 'iDOS [electrons]'
    else:
        ylabel = 'DOS [electrons/eV]'

    if plot:
        ax = dosobjs.plot()
        ax.set_xlabel(r'$\epsilon-\epsilon_F$ [eV]')
        ax.set_ylabel(ylabel)
        import matplotlib.pyplot as plt
        plt.show()



from ase import Atoms
from gpaw import GPAW
from gpaw.test import equal


def test_generic_Cl_minus():
    s = Atoms('Cl')
    s.center(vacuum=3)
    c = GPAW(mode='fd',
             xc={'name': 'PBE', 'stencil': 1},
             nbands=-4,
             charge=-1,
             h=0.3)
    s.calc = c

    e = s.get_potential_energy()
    niter = c.get_number_of_iterations()

    print(e, niter)
    energy_tolerance = 0.004
    equal(e, -2.8967, energy_tolerance)



"""Simple magnons calculation in a 1D hydrogen chain."""

# General modules
import pytest
import numpy as np

# Script modules
from ase import Atoms
from ase.dft.kpoints import monkhorst_pack

from gpaw import PW, GPAW
from gpaw.mpi import world
from gpaw.test import findpeak

from gpaw.response import ResponseGroundStateAdapter
from gpaw.response.frequencies import ComplexFrequencyDescriptor
from gpaw.response.chiks import ChiKSCalculator
from gpaw.response.susceptibility import ChiFactory
from gpaw.response.fxc_kernels import AdiabaticFXCCalculator
from gpaw.response.dyson import HXCScaling
from gpaw.response.pair_functions import read_pair_function


@pytest.mark.later  # interpolate=3 for PW-mode not implemented!
@pytest.mark.kspair
@pytest.mark.response
def test_response_afm_hchain_gssALDA(in_tmp_dir):
    # ---------- Inputs ---------- #

    # Part 1: Ground state calculation
    # Define atomic structure
    a = 2.5
    vfactor = 0.8
    mm = 1.
    # Ground state calculator options
    xc = 'LDA'
    kpts = 12
    nbands = 2 * (1 + 0)  # 1s + 0 empty shell bands
    ebands = 2 * 1  # Include also 2s bands for numerical consistency
    pw = 250
    conv = {'bands': nbands}

    # # Part 2: Magnetic response calculation
    q_qc = [[0., 0., 0.],
            [1. / 6., 0., 0.],
            [1. / 3., 0., 0.]]
    fxc = 'ALDA'
    hxc_scaling = HXCScaling('afm')
    rshelmax = -1
    rshewmin = 1e-8
    ecut = 120
    frq_w = np.linspace(-0.6, 0.6, 41)
    eta = 0.24
    zd = ComplexFrequencyDescriptor.from_array(frq_w + 1.j * eta)
    if world.size % 4 == 0:
        nblocks = 4
    elif world.size % 2 == 0:
        nblocks = 2
    else:
        nblocks = 1

    # ---------- Script ---------- #

    # Part 1: Ground state calculation

    Hatom = Atoms('H',
                  cell=[a, 0, 0],
                  # Use pbc to allow for real-space density interpolation
                  pbc=[1, 1, 1])
    Hatom.center(vacuum=vfactor * a, axis=(1, 2))
    Hchain = Hatom.repeat((2, 1, 1))
    Hchain.set_initial_magnetic_moments([mm, -mm])

    calc = GPAW(xc=xc,
                mode=PW(pw,
                        # Interpolate the density in real space
                        interpolation=3),
                kpts=monkhorst_pack((kpts, 1, 1)),
                nbands=nbands + ebands,
                convergence=conv,
                symmetry={'point_group': True},
                parallel={'domain': 1})

    Hchain.calc = calc
    Hchain.get_potential_energy()

    # Part 2: Magnetic response calculation
    gs = ResponseGroundStateAdapter(calc)
    chiks_calc = ChiKSCalculator(gs,
                                 nbands=nbands,
                                 ecut=ecut,
                                 gammacentered=True,
                                 nblocks=nblocks)
    fxc_calculator = AdiabaticFXCCalculator.from_rshe_parameters(
        gs, chiks_calc.context,
        rshelmax=rshelmax,
        rshewmin=rshewmin)
    chi_factory = ChiFactory(chiks_calc, fxc_calculator)

    for q, q_c in enumerate(q_qc):
        filename = 'h-chain_macro_tms_q%d.csv' % q
        txt = 'h-chain_macro_tms_q%d.txt' % q
        _, chi = chi_factory('+-', q_c, zd,
                             fxc=fxc,
                             hxc_scaling=hxc_scaling,
                             txt=txt)
        chi.write_macroscopic_component(filename)

    chi_factory.context.write_timer()
    world.barrier()

    # Part 3: Identify magnon peak in finite q scattering function
    w0_w, chi0_w = read_pair_function('h-chain_macro_tms_q0.csv')
    w1_w, chi1_w = read_pair_function('h-chain_macro_tms_q1.csv')
    w2_w, chi2_w = read_pair_function('h-chain_macro_tms_q2.csv')

    wpeak1, Ipeak1 = findpeak(w1_w, -chi1_w.imag / np.pi)
    wpeak2, Ipeak2 = findpeak(w2_w, -chi2_w.imag / np.pi)
    mw1 = calculate_afm_mw(wpeak1, eta) * 1000  # to meV
    mw2 = calculate_afm_mw(wpeak2, eta) * 1000  # to meV

    # Part 4: compare new results to test values
    test_fxcs = 1.04744
    test_mw1 = 285.  # meV
    test_mw2 = 494.  # meV  # Remark that mw2 < 2 * mw1 (linear dispersion)
    test_Ipeak1 = 0.0131  # a.u.
    test_Ipeak2 = 0.0290

    # Test fxc_scaling:
    fxcs = hxc_scaling.lambd
    assert abs(fxcs - test_fxcs) < 0.005

    # Magnon peak at q=1/3 q_X:
    assert abs(mw1 - test_mw1) < 10.

    # Magnon peak at q=2/3 q_X:
    assert abs(mw2 - test_mw2) < 10.

    # Scattering function intensity:
    assert abs(Ipeak1 - test_Ipeak1) < 0.005
    assert abs(Ipeak2 - test_Ipeak2) < 0.005

    # Check that spectrum at q=0 vanishes
    chitol = 1e-3 * np.abs(chi1_w.imag).max()
    assert np.abs(chi0_w.imag).max() < chitol

    # Check that the spectrum is antisymmetric around q=0
    assert np.allclose(w0_w[19::-1] + w0_w[21:], 0.)
    assert np.allclose(chi0_w.imag[19::-1] + chi0_w.imag[21:], 0.,
                       atol=0.01 * chitol)
    assert np.allclose(chi1_w.imag[19::-1] + chi1_w.imag[21:], 0.,
                       atol=chitol)
    assert np.allclose(chi2_w.imag[19::-1] + chi2_w.imag[21:], 0.,
                       atol=chitol)


# ---------- Script functionality ---------- #


def calculate_afm_mw(peak_frequency, eta):
    """Assume lorentzian lineshape to compute the magnon frequency
    from the afm magnon lineshape peak position."""
    return np.sqrt(2 * np.sqrt(peak_frequency**4 + eta**2 * peak_frequency**2)
                   - peak_frequency**2 - eta**2)



import pytest
import numpy as np
from gpaw.mpi import world, send, receive, broadcast_array


def test_send_receive_object():
    if world.size == 1:
        return
    obj = (42, 'hello')
    if world.rank == 0:
        send(obj, 1, world)
    elif world.rank == 1:
        assert obj == receive(0, world)


@pytest.mark.ci
def test_scalar_reduce():
    assert world.sum_scalar(world.rank + 1) == world.size * \
           (world.size + 1) // 2
    assert np.allclose(world.sum_scalar(world.rank + 1.0),
                       world.size * (world.size + 1.0) / 2)
    assert world.min_scalar(world.rank + 1) == 1
    assert world.max_scalar(world.rank + 1) == world.size
    assert world.min_scalar(world.rank + 1.0) == 1.0
    assert world.max_scalar(world.rank + 1.0) == world.size * 1.0
    

def test_bcast_array():
    new = world.new_communicator

    if world.size == 2:
        comms = [world]
    elif world.size == 4:
        ranks = np.array([[0, 1], [2, 3]])
        comms = [new(ranks[world.rank // 2]),
                 new(ranks[:, world.rank % 2])]
    elif world.size == 8:
        ranks = [[[0, 1], [0, 2], [0, 4]],
                 [[0, 1], [1, 3], [1, 5]],
                 [[2, 3], [0, 2], [2, 6]],
                 [[2, 3], [1, 3], [3, 7]],
                 [[4, 5], [4, 6], [0, 4]],
                 [[4, 5], [5, 7], [1, 5]],
                 [[6, 7], [4, 6], [2, 6]],
                 [[6, 7], [5, 7], [3, 7]]]
        comms = [new(r) for r in ranks[world.rank]]
    else:
        return

    array = np.zeros(3, int)
    if world.rank == 0:
        array[:] = 42

    out = broadcast_array(array, *comms)
    assert (out == 42).all()



import pytest
from ase.build import molecule

from gpaw import GPAW
from gpaw.atom.basis import BasisMaker
from gpaw.poisson import FDPoissonSolver as PoissonSolver
from gpaw.test import equal


@pytest.mark.later
def test_lcao_bsse():
    """Tests basis set super position error correction.

    Compares a single hydrogen atom to a system of one hydrogen atom
    and one ghost hydrogen atom.  The systems should have identical properties,
    i.e. the ghost orbital should have a coefficient of 0.
    """

    b = BasisMaker('H').generate(1, 0, energysplit=0.005)

    system = molecule('H2')
    system.center(vacuum=6.0)

    def prepare(setups):
        calc = GPAW(basis={'H': b}, mode='lcao',
                    setups=setups, h=0.2,
                    poissonsolver=PoissonSolver(nn='M', relax='GS', eps=1e-5),
                    spinpol=False,
                    nbands=1)
        system.calc = calc
        return calc

    calc = prepare({0: 'paw', 1: 'ghost'})
    system.calc = calc
    e_bsse = system.get_potential_energy()
    niter_bsse = calc.get_number_of_iterations()

    c_nM = calc.wfs.kpt_u[0].C_nM
    print('coefs')
    print(c_nM)
    print('energy', e_bsse)

    # Reference system which is just a hydrogen
    sys0 = system[0:1].copy()
    calc = prepare('paw')
    sys0.calc = calc
    e0 = sys0.get_potential_energy()
    niter0 = calc.get_number_of_iterations()
    print('e0, e_bsse = ', e0, e_bsse)

    # One coefficient should be very small (0.012), the other very large (0.99)
    assert abs(1.0 - abs(c_nM[0, 0])) < 0.02
    assert abs(c_nM[0, 1]) < 0.02
    assert abs(e_bsse - e0) < 2e-3

    energy_tolerance = 0.002
    niter_tolerance = 3
    equal(e_bsse, 0.02914, energy_tolerance)  # svnversion 5252
    equal(niter_bsse, 7, niter_tolerance)  # svnversion 5252
    equal(e0, 0.03038, energy_tolerance)  # svnversion 5252
    equal(niter0, 6, niter_tolerance)  # svnversion 5252



import pytest
from gpaw import GPAW, FermiDirac
from gpaw.mpi import world, size, rank
from gpaw.lrtddft2 import LrTDDFT2
from gpaw.lrtddft2.lr_communicators import LrCommunicators
from gpaw.test import equal
from ase.atoms import Atoms


@pytest.mark.lrtddft
def test_lrtddft2_Al2(in_tmp_dir):
    debug = False
    restart_file = 'Al2_gs.gpw'

    d = 2.563
    atoms = Atoms('Al2', positions=((0, 0, 0),
                                    (0, 0, d)))
    atoms.center(4.0)
    calc = GPAW(mode='fd', h=0.24, eigensolver='cg', basis='dzp',
                occupations=FermiDirac(width=0.01),
                convergence={'eigenstates': 4.0e-5,
                             'density': 1.0e-2,
                             'bands': 'all'},
                nbands=20)
    atoms.calc = calc
    atoms.get_potential_energy()
    calc.write(restart_file, mode='all')

    # Try to run parallel over eh-pairs
    if size % 2 == 0:
        eh_size = 2
        domain_size = size // eh_size
    else:
        eh_size = 1
        domain_size = size

    lr_comms = LrCommunicators(world, domain_size, eh_size)

    calc = GPAW(restart_file,
                communicator=lr_comms.dd_comm)
    de = 3.0
    lr = LrTDDFT2('Al2_lri',
                  calc,
                  fxc='PBE',
                  max_energy_diff=de,
                  lr_communicators=lr_comms
                  )
    w, S, R, Sx, Sy, Sz = lr.get_transitions(max_energy=1e9, units='au')

    e0_1 = w[0]
    e1_1 = w[-1]
    # Continue with larger de
    de = 4.5
    lr = LrTDDFT2('Al2_lri',
                  calc,
                  fxc='PBE',
                  max_energy_diff=de,
                  lr_communicators=lr_comms)
    w, S, R, Sx, Sy, Sz = lr.get_transitions(max_energy=1e9, units='au')
    e0_2 = w[0]
    e1_2 = w[-1]
    # Continue with smaller de
    de = 2.5
    lr = LrTDDFT2('Al2_lri',
                  calc,
                  fxc='PBE',
                  max_energy_diff=de,
                  lr_communicators=lr_comms)
    w, S, R, Sx, Sy, Sz = lr.get_transitions(max_energy=1e9, units='au')
    e0_3 = w[0]
    e1_3 = w[-1]

    if debug and rank == 0:
        print(e0_1, e1_1)
        print(e0_2, e1_2)
        print(e0_3, e1_3)

    tol = 5.0e-8
    equal(e0_1, e0_2, tol)
    equal(e0_1, e0_3, tol)
    tol = 1.0e-3
    equal(e0_1, 0.00105074187176, tol)
    equal(e1_1, 0.183188157301, tol)
    equal(e1_2, 0.194973135812, tol)
    equal(e1_3, 0.120681529342, tol)



import numpy as np
import pytest
from ase import Atoms
from gpaw import PW, FermiDirac
from gpaw.new.ase_interface import GPAW as NewGPAW
from gpaw import GPAW as AnyGPAW, SCIPY_VERSION
from gpaw.mpi import world

# Domain and k-point parallelization:
dk = []
for d in [1, 2, 4, 8]:
    for k in [1, 2]:
        if d * k > world.size:
            continue
        dk.append((d, k))


@pytest.mark.stress
@pytest.mark.parametrize('d, k', dk)
@pytest.mark.parametrize(
    'gpu', [False,
            pytest.param(
                True,
                marks=[pytest.mark.gpu,
                       pytest.mark.skipif(SCIPY_VERSION < [1, 6],
                                          reason='Too old scipy')])])
def test_pw_par_strategies(in_tmp_dir, d, k, gpu, gpaw_new):
    if (gpu or gpaw_new) and d * k < world.size:
        pytest.skip()
    ecut = 200
    kpoints = [1, 1, 4]
    atoms = Atoms('HLi',
                  cell=[6, 6, 3.4],
                  pbc=True,
                  positions=[[3, 3, 0],
                             [3, 3, 1.6]])
    parallel = {'domain': d, 'kpt': k}
    if gpu:
        parallel['gpu'] = True
        GPAW = NewGPAW
    else:
        GPAW = AnyGPAW
    atoms.calc = GPAW(mode=PW(ecut),
                      txt='hli.txt',
                      parallel=parallel,
                      kpts={'size': kpoints},
                      convergence={'maximum iterations': 4},
                      occupations=FermiDirac(width=0.1))

    e = atoms.get_potential_energy()
    assert e == pytest.approx(-5.218064604018109, abs=1e-11)

    f = atoms.get_forces()
    assert f == pytest.approx(np.array([[0, 0, -7.85130336e-01],
                                        [0, 0, 8.00667631e-01]]))

    if not gpu and not gpaw_new:
        s = atoms.get_stress()
        assert s == pytest.approx(
            [3.98105501e-03, 3.98105501e-03, -4.98044912e-03, 0, 0, 0])

    if not gpu:
        atoms.calc.write('hli.gpw', mode='all')
        GPAW('hli.gpw', txt=None)



from collections.abc import Sequence
import numpy as np
from gpaw.utilities.blas import gemmdot
from gpaw.grid_descriptor import GridDescriptor
from gpaw.projections import Projections


class IBZ2BZMaps(Sequence):
    """Sequence of data maps from k-points in the IBZ to the full BZ.

    The sequence is indexed by the BZ index K."""

    def __init__(self, kd, spos_ac, R_asii, N_c):
        """Construct the IBZ2BZMapper.

        Parameters
        ----------
        kd : KPointDescriptor
        spos_ac : np.array
            Scaled atomic positions
        R_asii : list
            Rotations of the PAW projections under symmetry transformations
        N_c : iterable of 3 ints
            Number of grid points along axes
        """
        self.kd = kd
        self.spos_ac = spos_ac
        self.R_asii = R_asii

        # Scaled coordinates of the real-space grid
        self.r_cR = np.array(np.meshgrid(*[np.linspace(0, 1, N, endpoint=False)
                                           for N in N_c], indexing='ij'))

    @classmethod
    def from_calculator(cls, calc):
        R_asii = calc.setups.atomrotations.get_R_asii()
        return cls(calc.wfs.kd, calc.spos_ac, R_asii, calc.wfs.gd.N_c)

    def __len__(self):
        return len(self.kd.bzk_kc)

    def __getitem__(self, K):
        kd = self.kd
        return IBZ2BZMap(kd.ibzk_kc[kd.bz2ibz_k[K]],
                         *self.get_symmetry_transformations(kd.sym_k[K]),
                         kd.time_reversal_k[K],
                         self.spos_ac,
                         self.kd.bzk_kc[K],
                         self.r_cR)

    def get_symmetry_transformations(self, s):
        return (self.get_rotation_matrix(s),
                self.get_atomic_permutations(s),
                self.get_projections_rotations(s))

    def get_rotation_matrix(self, s):
        """Coordinate rotation matrix, mapping IBZ -> K."""
        U_cc = self.kd.symmetry.op_scc[s]
        return U_cc

    def get_atomic_permutations(self, s):
        """Permutations of atomic indices in the IBZ -> K map."""
        b_a = self.kd.symmetry.a_sa[s]  # Atom a is mapped onto atom b
        return b_a

    def get_projections_rotations(self, s):
        """Rotations of the PAW projections for the IBZ -> K mapping."""
        R_aii = [R_sii[s] for R_sii in self.R_asii]
        return R_aii


class IBZ2BZMap:
    """Functionality to map orbitals from the IBZ to a specific k-point K."""

    def __init__(self, ik_c, U_cc, b_a, R_aii,
                 time_reversal, spos_ac, k_c, r_cR):
        """Construct the IBZ2BZMap."""
        self.ik_c = ik_c
        self.k_c = k_c  # k_c in 1:st BZ that IBZ-kpoint is mapped to
        self.r_cR = r_cR
        self.U_cc = U_cc
        self.b_a = b_a
        self.R_aii = R_aii
        self.time_reversal = time_reversal

        self.spos_ac = spos_ac

    def map_kpoint(self):
        """Get the relative k-point coordinates after the IBZ -> K mapping.

        NB: The mapped k-point can lie outside the BZ, but will always be
        related to kd.bzk_kc[K] by a reciprocal lattice vector.
        """
        # Apply symmetry operations to the irreducible k-point
        sign = 1 - 2 * self.time_reversal
        k_c = sign * self.U_cc @ self.ik_c

        return k_c

    def map_pseudo_wave(self, ut_R):
        """Map the periodic part of the pseudo wave from the IBZ -> K.

        For the symmetry operator U, which maps K = U ik, where ik is the
        irreducible BZ k-point and K does not necessarily lie within the 1BZ,

        psit_K(r) = psit_ik(U^T r)

        The mapping takes place on the coarse real-space grid.
        """
        # Apply symmetry operations to the periodic part of the pseudo wave
        if not (self.U_cc == np.eye(3)).all():
            N_c = ut_R.shape
            i_cr = np.dot(self.U_cc.T, np.indices(N_c).reshape((3, -1)))
            i = np.ravel_multi_index(i_cr, N_c, 'wrap')
            utout_R = ut_R.ravel()[i].reshape(N_c)
        else:
            utout_R = ut_R.copy()
        if self.time_reversal:
            utout_R = utout_R.conj()

        assert utout_R is not ut_R, \
            "We don't want the output array to point back at the input array"

        return utout_R

    def map_pseudo_wave_to_BZ(self, ut_R):
        """Map the periodic part of wave function from IBZ -> k in BZ.

        Parameters
        ----------
        ut_R: np.array
              Periodic part of pseudo wf at IBZ k-point
        """
        # Map the pseudo wave and K-point using symmetry operations
        utout_R = self.map_pseudo_wave(ut_R)
        kpt_shift_c = self.map_kpoint() - self.k_c

        # Check if the resulting K-point already resides inside the BZ
        if np.allclose(kpt_shift_c, 0.0):
            return utout_R

        # Check that the mapped k-point differ from the BZ K-point by
        # a reciprocal lattice vector G
        assert np.allclose((kpt_shift_c - np.round(kpt_shift_c)), 0.0)

        # Add a phase e^iG.r to the periodic part of the wf
        return utout_R * np.exp(2j * np.pi * gemmdot(kpt_shift_c, self.r_cR))

    def map_projections(self, projections):
        """Perform IBZ -> K mapping of the PAW projections.

        The projections of atom "a" are mapped onto an atom related to atom
        "b" by a lattice vector:

        r_b = U^T r_a    (or equivalently: r_b^T = r_a^T U)

        This means that when mapping

        psi_K(r) = psi_ik(U^T r),

        we need to generate the projections at atom a for k-point K based on
        the projections at atom b for k-point ik.
        """
        mapped_projections = projections.new()
        if mapped_projections.atom_partition.comm.rank == 0:
            for a, (b, U_ii) in enumerate(zip(self.b_a, self.U_aii)):
                # Map projections
                Pin_ni = projections[b]
                Pout_ni = Pin_ni @ U_ii
                if self.time_reversal:
                    Pout_ni = np.conj(Pout_ni)

                # Store output projections
                I1, I2 = mapped_projections.map[a]
                mapped_projections.array[..., I1:I2] = Pout_ni
        else:
            assert len(mapped_projections.indices) == 0

        return mapped_projections

    @property
    def U_aii(self):
        """Phase corrected rotation matrices for the PAW projections."""
        U_aii = []
        for a, R_ii in enumerate(self.R_aii):
            # The symmetry transformation maps atom "a" to a position which is
            # related to atom "b" by a lattice vector (but which does not
            # necessarily lie within the unit cell)
            b = self.b_a[a]
            cell_shift_c = self.spos_ac[a] @ self.U_cc - self.spos_ac[b]
            assert np.allclose(cell_shift_c.round(), cell_shift_c)
            # This means, that when we want to extract the projections at K for
            # atom a according to psi_K(r_a) = psi_ik(U^T r_a), we need the
            # projections at U^T r_a for k-point ik. Since we only have the
            # projections within the unit cell we need to multiply them with a
            # phase factor according to the cell shift.
            phase_factor = np.exp(2j * np.pi * self.ik_c @ cell_shift_c)
            U_ii = R_ii.T * phase_factor
            U_aii.append(U_ii)

        return U_aii


def get_overlap(bands,
                gd: GridDescriptor,
                ut1_nR, ut2_nR,
                proj1: Projections,
                proj2: Projections,
                dO_aii):
    """Computes the overlap between all-electron wave functions.

    Similar to gpaw.berryphase.get_overlap but adapted to work with projector
    objects rather than arrays.
    Eventually berryphase.get_overlap should be replaced by this function. XXX

    Parameters
    ----------
    bands:  list of integers
        Band indices to calculate overlap for
    dO_aii: dict
        Overlap coefficients for the partial waves in the PAW setups.
        Can be extracted via get_overlap_coefficients().
    """
    return _get_overlap(bands, gd.dv, ut1_nR, ut2_nR, proj1, proj2, dO_aii)


def _get_overlap(bands, dv, ut1_nR, ut2_nR, proj1, proj2, dO_aii):
    nbands = len(bands)
    ut1_nR = np.reshape(ut1_nR[bands], (nbands, -1))
    ut2_nR = np.reshape(ut2_nR[bands], (nbands, -1))
    M_nn = (ut1_nR.conj() @ ut2_nR.T) * dv

    for a in proj1.map:
        dO_ii = dO_aii[a]
        if proj1.collinear:
            P1_ni = proj1[a][bands]
            P2_ni = proj2[a][bands]
            M_nn += P1_ni.conj() @ (dO_ii) @ (P2_ni.T)
        else:
            # We have an additional spinor index s to sum over
            P1_nsi = proj1[a][bands]
            P2_nsi = proj2[a][bands]
            assert P1_nsi.shape[1] == 2
            for s in range(2):
                M_nn += P1_nsi[:, s].conj() @ (dO_ii) @ (P2_nsi[:, s].T)

    return M_nn


def get_overlap_coefficients(wfs):
    dO_aii = {}
    for a in wfs.kpt_u[0].projections.map:
        dO_aii[a] = wfs.setups[a].dO_ii
    return dO_aii


def get_phase_shifted_overlap_coefficients(dO_aii, spos_ac, K_c):
    """Apply phase shift to overlap coefficients.

    Applies a Bloch phase factor e^(iK.R_a) to the overlap coefficients dO_aii
    at each (scaled) atomic position spos_ac (R_a) for the input (scaled) wave
    vector K_c.
    """
    new_dO_aii = {}
    for a, dO_ii in dO_aii.items():
        phase_factor = np.exp(2j * np.pi * K_c @ spos_ac[a])
        new_dO_aii[a] = phase_factor * dO_ii
    return new_dO_aii



# flake8: noqa
import time
import numpy as np
from gpaw.utilities.gauss import gaussian_wave


def test_gauss_wave(rng):
    sigma = 2.4
    C = 3
    G = 50

    t = time.time()
    r_cG = rng.normal(size=C * G**3).reshape((C, G, G, G))
    r0_c = rng.normal(size=C)
    k_c = rng.normal(size=C)
    A = rng.uniform() * np.exp(1j * rng.uniform(0, 2 * np.pi))
    print('Allocation: %8.5f s' % (time.time() - t))

    # -------------------------------------------------------------------

    # Test case for real-part of gamma-point wave with normalized amplitude
    _gaussRGN = lambda r_cG, r0_c, sigma: 1/(sigma*np.pi**0.5)**1.5 \
        * np.exp(-np.sum((r_cG-r0_c[:,np.newaxis,np.newaxis,np.newaxis])**2, axis=0)/(2*sigma**2))

    t = time.time()
    gs0_G = _gaussRGN(r_cG, r0_c, sigma)
    print('_gaussRGN: %8.5f s' % (time.time() - t))

    t = time.time()
    gs1_G = gaussian_wave(r_cG, r0_c, sigma)
    print('+gaussRGN: %8.5f s' % (time.time() - t))

    assert np.abs(gs0_G -
                  gs1_G).max() < 1e-12, 'Max error %g' % np.abs(gs0_G -
                                                                gs1_G).max()
    del gs0_G, gs1_G

    # Test case for real-part of gamma-point wave with complex amplitude
    _gaussRGA = lambda r_cG, r0_c, sigma, A: np.real(A) \
        * np.exp(-np.sum((r_cG-r0_c[:,np.newaxis,np.newaxis,np.newaxis])**2, axis=0)/(2*sigma**2))

    t = time.time()
    gs0_G = _gaussRGA(r_cG, r0_c, sigma, A)
    print('_gaussRGA: %8.5f s' % (time.time() - t))

    t = time.time()
    gs1_G = gaussian_wave(r_cG, r0_c, sigma, None, A)
    print('+gaussRGA: %8.5f s' % (time.time() - t))

    assert np.abs(gs0_G -
                  gs1_G).max() < 1e-12, 'Max error %g' % np.abs(gs0_G -
                                                                gs1_G).max()
    del gs0_G, gs1_G

    # Test case for real-part of kpoint-point wave with normalized amplitude
    _gaussRKN = lambda r_cG, r0_c, sigma, k_c: 1/(sigma*np.pi**0.5)**1.5 \
        * np.exp(-np.sum((r_cG-r0_c[:,np.newaxis,np.newaxis,np.newaxis])**2, axis=0)/(2*sigma**2)) \
        * np.cos(np.sum(r_cG*k_c[:,np.newaxis,np.newaxis,np.newaxis], axis=0))

    t = time.time()
    gs0_G = _gaussRKN(r_cG, r0_c, sigma, k_c)
    print('_gaussRKN: %8.5f s' % (time.time() - t))

    t = time.time()
    gs1_G = gaussian_wave(r_cG, r0_c, sigma, k_c)
    print('+gaussRKN: %8.5f s' % (time.time() - t))

    assert np.abs(gs0_G -
                  gs1_G).max() < 1e-12, 'Max error %g' % np.abs(gs0_G -
                                                                gs1_G).max()
    del gs0_G, gs1_G

    # Test case for real-part of kpoint-point wave with complex amplitude
    _gaussRKA = lambda r_cG, r0_c, sigma, k_c, A: \
        np.exp(-np.sum((r_cG-r0_c[:,np.newaxis,np.newaxis,np.newaxis])**2, axis=0)/(2*sigma**2)) \
        * np.real(A*np.exp(1j*np.sum(r_cG*k_c[:,np.newaxis,np.newaxis,np.newaxis], axis=0)))

    t = time.time()
    gs0_G = _gaussRKA(r_cG, r0_c, sigma, k_c, A)
    print('_gaussRKA: %8.5f s' % (time.time() - t))

    t = time.time()
    gs1_G = gaussian_wave(r_cG, r0_c, sigma, k_c, A)
    print('+gaussRKA: %8.5f s' % (time.time() - t))

    assert np.abs(gs0_G -
                  gs1_G).max() < 1e-12, 'Max error %g' % np.abs(gs0_G -
                                                                gs1_G).max()
    del gs0_G, gs1_G

    # -------------------------------------------------------------------

    # Test case for complex case of gamma-point wave with normalized amplitude
    _gaussCGN = lambda r_cG, r0_c, sigma: (1+0j)/(sigma*np.pi**0.5)**1.5 \
        * np.exp(-np.sum((r_cG-r0_c[:,np.newaxis,np.newaxis,np.newaxis])**2, axis=0)/(2*sigma**2))

    t = time.time()
    gs0_G = _gaussCGN(r_cG, r0_c, sigma)
    print('_gaussCGN: %8.5f s' % (time.time() - t))

    t = time.time()
    gs1_G = gaussian_wave(r_cG, r0_c, sigma, dtype=complex)
    print('+gaussCGN: %8.5f s' % (time.time() - t))

    assert np.abs(gs0_G -
                  gs1_G).max() < 1e-12, 'Max error %g' % np.abs(gs0_G -
                                                                gs1_G).max()
    del gs0_G, gs1_G

    # Test case for complex case of gamma-point wave with complex amplitude
    _gaussCGA = lambda r_cG, r0_c, sigma, A: A \
        * np.exp(-np.sum((r_cG-r0_c[:,np.newaxis,np.newaxis,np.newaxis])**2, axis=0)/(2*sigma**2))

    t = time.time()
    gs0_G = _gaussCGA(r_cG, r0_c, sigma, A)
    print('_gaussCGA: %8.5f s' % (time.time() - t))

    t = time.time()
    gs1_G = gaussian_wave(r_cG, r0_c, sigma, None, A, dtype=complex)
    print('+gaussCGA: %8.5f s' % (time.time() - t))

    assert np.abs(gs0_G -
                  gs1_G).max() < 1e-12, 'Max error %g' % np.abs(gs0_G -
                                                                gs1_G).max()
    del gs0_G, gs1_G

    # Test case for complex case of kpoint-point wave with normalized amplitude
    _gaussCKN = lambda r_cG, r0_c, sigma, k_c: (1+0j)/(sigma*np.pi**0.5)**1.5 \
        * np.exp(-np.sum((r_cG-r0_c[:,np.newaxis,np.newaxis,np.newaxis])**2, axis=0)/(2*sigma**2)) \
        * np.exp(1j*np.sum(r_cG*k_c[:,np.newaxis,np.newaxis,np.newaxis], axis=0))

    t = time.time()
    gs0_G = _gaussCKN(r_cG, r0_c, sigma, k_c)
    print('_gaussCKN: %8.5f s' % (time.time() - t))

    t = time.time()
    gs1_G = gaussian_wave(r_cG, r0_c, sigma, k_c, dtype=complex)
    print('+gaussCKN: %8.5f s' % (time.time() - t))

    assert np.abs(gs0_G -
                  gs1_G).max() < 1e-12, 'Max error %g' % np.abs(gs0_G -
                                                                gs1_G).max()
    del gs0_G, gs1_G

    # Test case for complex case of kpoint-point wave with complex amplitude
    _gaussCKA = lambda r_cG, r0_c, sigma, k_c, A: A \
        * np.exp(-np.sum((r_cG-r0_c[:,np.newaxis,np.newaxis,np.newaxis])**2, axis=0)/(2*sigma**2)) \
        * np.exp(1j*np.sum(r_cG*k_c[:,np.newaxis,np.newaxis,np.newaxis], axis=0))

    t = time.time()
    gs0_G = _gaussCKA(r_cG, r0_c, sigma, k_c, A)
    print('_gaussCKA: %8.5f s' % (time.time() - t))

    t = time.time()
    gs1_G = gaussian_wave(r_cG, r0_c, sigma, k_c, A, dtype=complex)
    print('+gaussCKA: %8.5f s' % (time.time() - t))

    assert np.abs(gs0_G -
                  gs1_G).max() < 1e-12, 'Max error %g' % np.abs(gs0_G -
                                                                gs1_G).max()
    del gs0_G, gs1_G



import pytest
from gpaw.mpi import world
from ase import Atoms
from gpaw import GPAW
from gpaw.test import equal

pytestmark = pytest.mark.skipif(world.size > 1,
                                reason='world.size > 1')


@pytest.mark.legacy
def test_potential():
    for mode in ['fd', 'pw']:
        print(mode)
        hydrogen = Atoms('H',
                         cell=(2.5, 3, 3.5),
                         pbc=1,
                         calculator=GPAW(txt=None, mode=mode))
        hydrogen.get_potential_energy()
        dens = hydrogen.calc.density
        ham = hydrogen.calc.hamiltonian
        wfs = hydrogen.calc.wfs
        ham.poisson.eps = 1e-20
        dens.interpolate_pseudo_density()
        dens.calculate_pseudo_charge()
        ham.update(dens)
        ham.get_energy(0.0, wfs)
        y = (ham.vt_sG[0, 0, 0, 0] - ham.vt_sG[0, 0, 0, 1]) * ham.gd.dv
        x = 0.0001
        dens.nt_sG[0, 0, 0, 0] += x
        dens.nt_sG[0, 0, 0, 1] -= x
        dens.interpolate_pseudo_density()
        dens.calculate_pseudo_charge()
        ham.update(dens)
        e1 = ham.get_energy(0.0, wfs) - ham.e_kinetic
        dens.nt_sG[0, 0, 0, 0] -= 2 * x
        dens.nt_sG[0, 0, 0, 1] += 2 * x
        dens.interpolate_pseudo_density()
        dens.calculate_pseudo_charge()
        ham.update(dens)
        e2 = ham.get_energy(0.0, wfs) - ham.e_kinetic
        equal(y, (e1 - e2) / (2 * x), 2e-8)



from gpaw.xc.gllb.nonlocalfunctional import NonLocalFunctional


def get_nonlocal_functional(name: str,
                            stencil: int = 2,
                            metallic: bool = False,
                            width: float = None,
                            eps: float = 0.05) -> NonLocalFunctional:
    """Function for building GLLB functionals.

    Parameters:

    name:
        name of the functional
    stencil:
        parameter passed to :class:`~gpaw.xc.gga.GGA`
    metallic:
        parameter passed to :class:`~gpaw.xc.gllb.coefficients.Coefficients`
    width:
        parameter passed to :class:`~gpaw.xc.gllb.coefficients.Coefficients`
    eps:
        parameter passed to :class:`~gpaw.xc.gllb.coefficients.Coefficients`

    Recognized names and implied parameters:
    * GLLB (Contains screening part from B88 functional
            and response part based on simple square root expression
            of orbital energy differences)
    * GLLBM (GLLB with `metallic=True`)
    * GLLBC (GLLB with screening part from PBE + correlation from PBE)
    * GLLBCP86 (GLLB with screening part from B88 + correlation from P86)
    * GLLBSC (GLLB with screening part from PBEsol + correlation from PBEsol)
    * GLLBSCM (GLLBSC with `metallic=True`)
    * GLLBNORESP (Just GLLB screening)
    * GLLBLDA (A test functional, which is just LDA but via
               NonLocalFunctional framework)
    * GLLBPBE (A test functional, which is just PBE but via
               NonLocalFunctional framework)
    """
    from gpaw.xc.gllb.c_gllbscr import C_GLLBScr
    from gpaw.xc.gllb.c_response import C_Response
    from gpaw.xc.gllb.c_xc import C_XC
    from gpaw.xc.gllb.coefficients import Coefficients

    functional = NonLocalFunctional(name)

    scr_functional = None
    xc_functional = None
    response = True
    setup_name = None

    # Set parameters based on the name
    if name == 'GLLB':
        scr_functional = 'GGA_X_B88'
    elif name == 'GLLBM':
        scr_functional = 'GGA_X_B88'
        setup_name = 'GLLB'
        metallic = True
    elif name == 'GLLBSC':
        scr_functional = 'GGA_X_PBE_SOL'
        xc_functional = 'GGA_C_PBE_SOL'
        setup_name = 'GLLBSC'
    elif name == 'GLLBSCM':
        scr_functional = 'GGA_X_PBE_SOL'
        xc_functional = 'GGA_C_PBE_SOL'
        setup_name = 'GLLBSC'
        metallic = True
    elif name == 'GLLBC':
        scr_functional = 'GGA_X_PBE'
        xc_functional = 'GGA_C_PBE'
    elif name == 'GLLBCP86':
        scr_functional = 'GGA_X_B88'
        xc_functional = 'GGA_C_P86'
    elif name == 'GLLBLDA':
        xc_functional = 'LDA'
        response = False
        setup_name = 'LDA'
    elif name == 'GLLBPBE':
        xc_functional = 'PBE'
        response = False
        setup_name = 'PBE'
    elif name == 'GLLBNORESP':
        scr_functional = 'GGA_X_B88'
        response = False
    else:
        raise RuntimeError('Unkown nonlocal density functional: ' + name)

    def func_name_to_dict(name):
        dct = {'name': name}
        if name != 'LDA':
            dct['stencil'] = stencil
        return dct

    # Construct functional
    functional = NonLocalFunctional(name, setup_name=setup_name)
    if scr_functional is not None:
        scr = C_GLLBScr(1.0, functional=func_name_to_dict(scr_functional))
        functional.add_contribution(scr)
    if response:
        coef = Coefficients(eps=eps, width=width)
        resp = C_Response(1.0, coef, metallic=metallic)
        functional.add_contribution(resp)
    if xc_functional is not None:
        xc = C_XC(1.0, functional=func_name_to_dict(xc_functional))
        functional.add_contribution(xc)
    return functional



from math import pi, sin, cos
import numpy as np

from ase.units import Bohr
import _gpaw
from gpaw.poisson import PoissonSolver
from gpaw.pes.state import State
from gpaw.analyse.expandyl import AngularIntegral


class PlaneWave(State):

    """Plane wave state."""

    def get_grid(self, k_c, r0=None):
        if r0 is None:
            r0_c = np.array([0., 0., 0.])
        else:
            r0_c = np.array(r0)
        pw_G = self.gd.zeros(dtype=complex)
        python = True
        python = False
        if python:
            r_c = np.array([0., 0., 0.])
            gd = self.gd
            for i in range(gd.n_c[0]):
                r_c[0] = (gd.beg_c[0] + i) * gd.h_c[0]
                for j in range(gd.n_c[1]):
                    r_c[1] = (gd.beg_c[1] + j) * gd.h_c[1]
                    for k in range(gd.n_c[2]):
                        r_c[2] = (gd.beg_c[2] + k) * gd.h_c[2]
                        phase = np.dot(k_c, r_c - r0_c)
                        pw_G[i, j, k] = cos(phase) + 1j * sin(phase)
        else:
            gd = self.gd
            _gpaw.plane_wave_grid(gd.beg_c, gd.end_c,
                                  gd.h_cv.diagonal().copy(),
                                  k_c, r0_c, pw_G)
        pw_G /= (2 * pi) ** (3. / 2.)
        return pw_G


class ZerothOrder1(State):

    def __init__(self, calculator):
        self.calculator = calculator
        State.__init__(self, calculator.gd)
        self.pw = PlaneWave(self.gd)

        # get effective potential
        hamiltonian = self.calculator.hamiltonian
        if 1:
            self.vt_G = hamiltonian.vt_sG[0]  # XXX treat spin
        else:
            self.vt_G = np.where(hamiltonian.vt_sG[0] > 0,
                                 0.0, hamiltonian.vt_sG[0])
        self.intvt = self.gd.integrate(self.vt_G)
        print('# int(vt_G)=', self.intvt, (self.vt_G > 0).any())

        self.solve()

    def get_grid(self, k_c, r0):
        print('Correction:', self.gd.integrate(self.corrt_G))
        if not hasattr(self, 'written'):
            print('r0', r0, r0 * Bohr)
            dR = 0.3
            AI = AngularIntegral(r0 * Bohr, self.gd, dR=dR)
            r_R = AI.radii()
            psi_R = AI.average(self.corrt_G)
            v_R = AI.average(self.vt_G)
            with open('radial_dR' + str(dR) + '.dat', 'w') as fd:
                print('# R  v(R)    psi(R)', file=fd)
                for r, psi, v in zip(r_R, psi_R, v_R):
                    print(r, psi, v, file=fd)

            self.written = True

        return self.pw.get_grid(k_c, r0) - 1e6 * self.corrt_G

    def solve(self):
        hamiltonian = self.calculator.hamiltonian
        self.poisson = PoissonSolver('fd', nn=hamiltonian.poisson.nn)
        self.poisson.set_grid_descriptor(self.gd)
        self.poisson.initialize()

        corrt_G = self.gd.empty()
        self.poisson.solve(corrt_G, self.vt_G, charge=None)
        corrt_G /= (2 * pi) ** (5. / 2)

        self.corrt_G = corrt_G
